import json

from django.core.cache import cache
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, Client

from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase


class CacheOps(Component):
    first_name: str = "Simon"

    def inline_template(self):
        return """
            <div>
                {{ first_name }}
            </div>            
            """


class OpsTestCase(SilicaTestCase):
    def setUp(self):
        self.client = Client()

    def test_can_see_cached_property(self):
        request = RequestFactory().get("/")
        request.user = AnonymousUser()

        response = CacheOps.as_view()(request)
        component_id = response.component.component_id

        # Simulate an attribute stored in cache
        cache.set(f"silica:component:{component_id}", {"first_name": "James"})

        response = self.client.post(
            "/silica/message",
            json.dumps(
                {
                    "id": component_id,
                    "name": f"{__name__}.CacheOps",
                }
            ),
            content_type="application/json",
        )

        self.assertContains(response, "James")

    def test_state_is_preserved_between_requests(self):
        request = RequestFactory().get("/")
        request.user = AnonymousUser()

        response = CacheOps.as_view()(request)
        component_id = response.component.component_id

        # Simulate an attribute stored in cache
        cache.set(f"silica:component:{component_id}", {"first_name": "James"})

        # Simulate a /message request
        response = self.client.post(
            "/silica/message",
            json.dumps(
                {
                    "id": component_id,
                    "name": f"{__name__}.CacheOps",
                }
            ),
            content_type="application/json",
        )

        self.assertContains(response, "James")

        # A subsequent request to /message should return the same state
        response = self.client.post(
            "/silica/message",
            json.dumps(
                {
                    "id": component_id,
                    "name": f"{__name__}.CacheOps",
                }
            ),
            content_type="application/json",
        )

        self.assertContains(response, "James")

    def test_cache_expired_exception_raised(self):
        pass
        # Simulate an attribute stored in cache
        # cache.set(f"silica:component:123", {"first_name": "James"})
        # with self.assertRaises(Exception):
        #     pass
            #call message function directly in order to catch exceptio
            #
            # self.client.post(
            #     "/silica/message",
            #     json.dumps(
            #         {
            #             "id": 123,
            #             "name": "silica.tests.components.CacheOps.CacheOps",
            #         }
            #     ),
            #     content_type="application/json",
            # )




from django.views.generic import TemplateView
from django.http import HttpResponse
from django.template import Context, Template


def create_test_view(inline_template_string, context_dict=None):
    """
    Dynamically creates a class-based view with the provided inline template string and context.

    :param inline_template_string: The template string to be rendered.
    :param context_dict: Optional dictionary to be used as context for rendering the template.
    :return: A class-based view dynamically generated.
    """
    if context_dict is None:
        context_dict = {}

    class DynamicTestView(TemplateView):
        def get(self, request, *args, **kwargs):
            context_dict['request'] = request

            template = Template("""
                <html>
                <head>
                {% load silica %}
                {% load silica_scripts %}
                {% silica_scripts %} 
                <link rel="icon" type="image/png" href="data:image/png;base64,iVBORw0KGgo=">
                </head>
                <body>   
            """ + inline_template_string + """
                </body>
                </html>
            """)
            context = Context(context_dict)
            rendered_template = template.render(context)
            return HttpResponse(rendered_template)

    return DynamicTestView


def data_get(data, path, default=None):
    """
    Retrieves a value from a nested data structure using "dot" notation.
    This function is designed to work with dictionaries, lists, tuples, and objects.
    It also supports keys with dots by using square bracket notation.

    Parameters
    ----------
    data : any
        The data structure from which to retrieve the value.
        This can be a nested structure of dictionaries, lists, tuples, and objects.
    path : str
        The path to the value to be retrieved. Use dot notation for nested structures.
        Use square brackets for keys that contain dots.
        Example: 'user.name[first.name]' would look for data['user']['name']['first.name']
    default : any, optional
        The default value to return if the specified path is not found in the data structure.
        By default, it is None.

    Returns
    -------
    any
        The value found at the specified path in the data structure.
        If the path is not found, the default value is returned.

    Examples
    --------
    # Nested dictionary example
    data = {'user': {'name': {'first.name': 'John', 'last': 'Doe'}}}
    print(data_get(data, 'user.name[first.name]'))  # Outputs: John

    # List example
    data = ['John', 'Doe']
    print(data_get(data, '0'))  # Outputs: John

    # Mixed example with dotted key
    data = {'users': [{'name.full': 'John Doe'}, {'name.full': 'Jane Doe'}]}
    print(data_get(data, 'users.1[name.full]'))  # Outputs: Jane Doe

    # Object example
    class User:
        def __init__(self, name):
            self.name = {'full.name': name}
    data = User('John Doe')
    print(data_get(data, 'name[full.name]'))  # Outputs: John Doe
    """

    def get_item(obj, key):
        if isinstance(obj, dict):
            return obj.get(key, default)
        elif isinstance(obj, (list, tuple)) and key.isdigit():
            index = int(key)
            return obj[index] if 0 <= index < len(obj) else default
        elif hasattr(obj, key):
            return getattr(obj, key)
        else:
            return default

    def parse_path(path):
        parts = []
        current = ''
        bracket_depth = 0
        for char in path:
            if char == '.' and bracket_depth == 0:
                if current:
                    parts.append(current)
                    current = ''
            elif char == '[':
                bracket_depth += 1
                if bracket_depth == 1 and current:
                    parts.append(current)
                    current = ''
                else:
                    current += char
            elif char == ']':
                bracket_depth -= 1
                if bracket_depth == 0:
                    parts.append(current)
                    current = ''
                else:
                    current += char
            else:
                current += char
        if current:
            parts.append(current)
        return parts

    parts = parse_path(path)
    for part in parts:
        if data is None:
            return default
        data = get_item(data, part)
    return data



from silica import Component


class ComponentInSubfolder(Component):
    def inline_template(self):
        return """
            <div>
                I'm component in subfolder!
            </div>
        """



import time

from django.test import override_settings
from django.urls import path
from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.SilicaLoadingTargetExceptComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("silica-loading-target-except", TestView.as_view()),
]


class SilicaLoadingTargetExceptComponent(Component):
    def specific_call(self):
        time.sleep(0.4)

    def generic_call(self):
        time.sleep(0.4)

    def inline_template(self):
        return """
            <div>
                <div silica:loading silica:target.except="generic_call">
                    I'm specific loading
                </div>
                
                <div silica:loading>
                    I'm generic loading
                </div>
                
                <button id="generic_call" silica:click="generic_call">Click me</button>
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class SilicaInitTestCase(SilicaBrowserTestCase):
    def test_target_except_modifier(self):
        self.selenium.get(self.live_server_url + '/silica-loading-target-except')

        self.assertDontSee("I'm specific loading")
        self.assertDontSee("I'm generic loading")

        self.selenium.find_element(By.ID, 'generic_call').click()
        time.sleep(0.1)
        
        self.assertDontSee("I'm specific loading")
        self.assertSee("I'm generic loading")




from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class ComponentWithHook(Component):
    prop1 = 'hello'
    updated_called = 'no'
    updated_prop1_called = 'no'
    another_prop_called = 'no'

    def updated(self, key, value):
        if key == 'prop1':
            self.updated_called = value

    def updated_prop1(self, value):
        self.updated_prop1_called = value

    def inline_template(self):
        return """
            <div>
                props1: {{ prop1 }}
                updated_called: {{ updated_called }}
                updated_prop1_called: {{ updated_prop1_called }}
            </div>
        """


class TestUpdatedHooks(SilicaTestCase):
    def test_updated_hook_is_called(self):
        (
            SilicaTest(component=ComponentWithHook)
            .assertSet('prop1', 'hello')
            .set('prop1', 'world')
            .assertSet('updated_called', 'world')
            .assertSet('updated_prop1_called', 'world')
            .assertSee('props1: world')
            .assertSee('updated_called: world')
            .assertSee('updated_prop1_called: world')
        )


from pathlib import Path

from django.test import TestCase
from django.core.management import call_command
from django.core.management.base import CommandError
from django.conf import settings

from silica.utils import pascal_to_snake

class SilicaMvCommandTestCase(TestCase):
    def setUp(self):
        self.files = []
        self.base_dir = Path(getattr(settings, 'BASE_DIR', None))
        self.main_app_name = 'example_project'

    def tearDown(self):
        for file in self.files:
            if file.exists():
                file.unlink()

    def create_temp_component(self, component_name, subfolders=[]):
        # Convert component name to snake_case for the template
        snake_component_name = pascal_to_snake(component_name)

        component_path = self.base_dir / self.main_app_name / 'silica' / Path(*subfolders) / f"{component_name}.py"
        template_path = self.base_dir / self.main_app_name / 'templates' / 'silica' / Path(*subfolders) / f"{snake_component_name}.html"
        component_path.parent.mkdir(parents=True, exist_ok=True)
        template_path.parent.mkdir(parents=True, exist_ok=True)
        component_path.touch()
        template_path.touch()
        self.files.extend([component_path, template_path])
        return component_path, template_path

    def test_move_component_success(self):
        existing_component_name = "ExistingComponent"
        new_component_name = "NewComponent"

        existing_component_path, existing_template_path = self.create_temp_component(existing_component_name)
        new_component_path = self.base_dir / self.main_app_name / 'silica' / f"{new_component_name}.py"
        new_template_path = self.base_dir / self.main_app_name / 'templates' / 'silica' / f"{pascal_to_snake(new_component_name)}.html"

        call_command('silica', 'mv', existing_component_name, new_component_name, "--silent")

        self.assertFalse(existing_component_path.exists(), "Old component file still exists.")
        self.assertFalse(existing_template_path.exists(), "Old template file still exists.")
        self.assertTrue(new_component_path.exists(), "New component file not found.")
        self.assertTrue(new_template_path.exists(), "New template file not found.")

    def test_move_component_invalid_existing_component(self):
        with self.assertRaises(CommandError):
            call_command('silica', 'mv', 'NonExistentComponent', 'NewComponent', "--silent")

    def test_move_component_to_existing_component(self):
        existing_component_name = "ExistingComponent"
        new_component_name = "NewComponent"

        self.create_temp_component(existing_component_name)
        self.create_temp_component(new_component_name)

        with self.assertRaises(CommandError):
            call_command('silica', 'mv', existing_component_name, new_component_name, "--silent")

    # Assuming the pascal_to_snake function is correctly converting the component names



def test_move_component_with_subfolders(self):
    existing_component_name = "folder1.folder2.ExistingComponent"
    new_component_name = "folder1.folder2.NewComponent"

    # Extract the final component name
    *_, final_new_component_name = new_component_name.split('.')

    existing_component_path, existing_template_path = self.create_temp_component("ExistingComponent", ['folder1', 'folder2'])

    new_component_path = self.base_dir / self.main_app_name / 'silica' / 'folder1' / 'folder2' / f"{final_new_component_name}.py"
    new_template_path = self.base_dir / self.main_app_name / 'templates' / 'silica' / 'folder1' / 'folder2' / f"{pascal_to_snake(final_new_component_name)}.html"

    # Debugging output
    print("Final new component path:", new_component_path)
    print("Final new template path:", new_template_path)

    # Ensure new files do not exist before command execution
    if new_component_path.exists():
        new_component_path.unlink()
    if new_template_path.exists():
        new_template_path.unlink()

    call_command('silica', 'mv', existing_component_name, new_component_name, "--silent")

    # Asserting the existence of new files
    self.assertFalse(existing_component_path.exists(), "Old component file still exists.")
    self.assertFalse(existing_template_path.exists(), "Old template file still exists.")
    self.assertTrue(new_component_path.exists(), "New component file not found.")
    self.assertTrue(new_template_path.exists(), "New template file not found.")



from django import template
from django.templatetags.static import static
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def silica_scripts():
    silica_js = static("silica.js")
    morphdom_js = static("morphdom.js")
    alpine_morph_js = static("alpine3.morph.min.js")
    alpine_js = static("alpine3.min.js")
    return mark_safe(f"""
        <script src="{morphdom_js}"></script>
        <script src="{silica_js}"></script>
        <script src="{alpine_morph_js}"></script>
        <script src="{alpine_js}" defer></script>
        <style>
            [silica\:loading], 
            [silica\:loading\.init], 
            [silica\:loading\.proceeding] {{ display: none }}</style>
    """)



from __future__ import annotations

import json

import urllib.parse

from django.template.response import TemplateResponse
from django.urls import set_urlconf
from django.conf import settings

from bs4 import BeautifulSoup
from bs4.formatter import HTMLFormatter

from silica.errors import SingleRootNodeError

app_urls = getattr(settings, "ROOT_URLCONF", None)


class SilicaTemplateResponse(TemplateResponse):
    def __init__(
            self,
            request,
            template,
            context=None,
            content_type=None,
            status=None,
            charset=None,
            using=None,
            headers=None,
            component: "Component" = None,
            init_js=False,
            lazy=False,
    ):
        super().__init__(
            template=template,
            request=request,
            context=context,
            content_type=content_type,
            status=status,
            charset=charset,
            using=using,
        )
        self.component = component
        self.init_js = init_js
        self.lazy = lazy

    def render(self):
        set_urlconf(app_urls)

        response = super().render()

        if not self.component or not self.component.component_id:
            return response

        template_string = response.content.decode("utf-8").strip()

        # Assuming there's only one root element in the template
        soup = BeautifulSoup(template_string, "html.parser")

        if len(soup.contents) != 1:
            raise SingleRootNodeError(
                f'Silica template for component "{self.component.__module__}" must include a single root element')

        root_node = soup.contents[0]

        root_node.attrs["silica:id"] = self.component.component_id
        root_node.attrs["silica:name"] = self.component.component_name

        if self.init_js and not self.component.lazy:
            snapshot = self.component.create_snapshot()
            snapshot_json = json.dumps(snapshot)
            snapshot_url_encoded = urllib.parse.quote(snapshot_json)
            root_node.attrs["silica:initial-data"] = snapshot_url_encoded

        if self.component.lazy:
            root_node.attrs["silica:lazy"] = "true"
            # root_node.attrs["silica:initial-data"] = "{}"

        self.component.has_rendered = True

        soup.smooth()
        rendered_template = soup.encode(formatter=UnsortedAttributes()).decode("utf-8")

        self.component.rendered(rendered_template)
        self.component.store_state()

        response.content = rendered_template

        return response


class UnsortedAttributes(HTMLFormatter):
    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v



from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class Methods(Component):
    test_value = 1

    fruit = "banana"

    def set_apple(self):
        self.fruit = "apple"

    def set_fruit(self, fruit=""):
        self.fruit = fruit

    def inline_template(self):
        return """
            <div>
                {{ fruit }}
                <a silica:click="set_apple()">set apple</a>
            </div>        
        """


class MethodsTestCase(SilicaTestCase):
    def test_method_calling_without_args(self):
        (
            SilicaTest(component=Methods)
            .assertSet("fruit", "banana")
            .assertSee("banana")
            .call("set_apple")
            .assertSet("fruit", "apple")
        )

    def test_method_calling_with_args(self):
        (
            SilicaTest(component=Methods)
            .call("set_fruit", "cherry")
            .assertSet("fruit", "cherry")
            .assertSee("cherry")
        )



from django.core.management.base import BaseCommand
from django.core.management import CommandError
from .silica_commands.create import Command as CreateCommand
from .silica_commands.mv import Command as MvCommand
from .silica_commands.rm import Command as RmCommand

class Command(BaseCommand):
    help = 'Wrapper command for silica subcommands'

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(dest='subcommand', help='Silica subcommands')

        # Create subcommand
        create_cmd = CreateCommand()
        create_parser = subparsers.add_parser('create', help='Create command')
        create_cmd.add_arguments(create_parser)

        # Mv subcommand
        mv_cmd = MvCommand()
        mv_parser = subparsers.add_parser('mv', help='Move command')
        mv_cmd.add_arguments(mv_parser)
        
        # Rm subcommand
        rm_cmd = RmCommand()
        rm_parser = subparsers.add_parser('rm', help='Remove command')
        rm_cmd.add_arguments(rm_parser)


    def handle(self, *args, **options):
        subcommand = options['subcommand']

        if subcommand == 'create':
            cmd = CreateCommand()
        elif subcommand == 'mv':
            cmd = MvCommand()
        elif subcommand == 'rm':
            cmd = RmCommand()
        else:
            raise CommandError('Unknown subcommand "%s"' % subcommand)

        # Delegate to the appropriate command
        cmd.handle(*args, **options)



import time
from typing import List, Dict

from django.test import override_settings
from django.urls import path

from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.CheckboxComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("checkboxes", TestView.as_view()),
]


class CheckboxComponent(Component):
    prop1: bool = True
    prop2: int = 1
    prop3 = None
    prop4: str = 'hello'
    prop5: List = ['item1']
    prop6: Dict = {
        'consent': True
    }
    prop7: Dict = {
        'items': ['foo']
    }
    prop8: List = []
    prop9: List = []

    def inline_template(self):
        return """
            <div>
                <p>
                    prop1: {{ prop1 }}
                    <input type="checkbox" silica:model="prop1" id="checkbox1">
                </p>
                
                <p>
                    prop2: {{ prop2 }}
                    <input type="checkbox" silica:model="prop2" id="checkbox2">
                </p>
                
                <p>
                    prop3: {{ prop3 }}            
                    <input type="checkbox" silica:model="prop3" id="checkbox3">
                </p>
                
                <p>
                    prop4: {{ prop4 }}
                    <input type="checkbox" silica:model="prop4" id="checkbox4">
                </p>
                
                <p>
                    prop5: [{% for item in prop5 %}{{ item }}{% endfor %}]
                    <input type="checkbox" silica:model="prop5" id="item1" value="item1">
                    <input type="checkbox" silica:model="prop5" id="another_item" value="another_item">
                </p>     
                
                <p>
                    prop6: [{{ prop6.consent }}]
                    <input type="checkbox" silica:model="prop6.consent" id="checkbox6">
                </p>                
                
                <p>
                    prop7: [{% for item in prop7.items %}{{ item }}{% endfor %}]
                    <input type="checkbox" silica:model="prop7.items" id="foo" value="foo">
                    <input type="checkbox" silica:model="prop7.items" id="bar" value="bar">
                </p>
                
                <p>
                    prop8: [{% for item in prop8.items %}{{ item }}{% endfor %}]
                    <input type="checkbox" silica:model="prop8" id="checkbox8">
                </p>  
                
                <p>
                    prop9: [{% for item in prop9.items %}{{ item }}{% endfor %}]
                    <input type="checkbox" silica:model="prop9" id="checkbox9" value="">
                </p>                
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class CheckboxesTestCase(SilicaBrowserTestCase):

    def test_boolean_checkboxes(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop1: True" in self.selenium.page_source)
        self.assertTrue("prop2: 1" in self.selenium.page_source)
        self.assertTrue("prop3: None" in self.selenium.page_source)
        self.assertTrue("prop4: hello" in self.selenium.page_source)
        self.assertTrue("prop5: [item1]" in self.selenium.page_source)

        # Checkbox 1
        self.selenium.find_element(By.ID, 'checkbox1').click()
        time.sleep(0.2)
        self.assertTrue("prop1: False" in self.selenium.page_source)

        # Checkbox 2
        self.selenium.find_element(By.ID, 'checkbox2').click()
        time.sleep(0.2)
        self.assertTrue("prop1: False" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox2').click()
        time.sleep(0.2)
        self.assertTrue("prop2: True" in self.selenium.page_source)

        # Checkbox 3
        self.selenium.find_element(By.ID, 'checkbox3').click()
        time.sleep(0.2)
        self.assertTrue("prop3: True" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox3').click()
        time.sleep(0.2)
        self.assertTrue("prop3: False" in self.selenium.page_source)

        # Checkbox 4
        self.selenium.find_element(By.ID, 'checkbox4').click()
        time.sleep(0.2)
        self.assertTrue("prop4: False" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox4').click()
        time.sleep(0.2)
        self.assertTrue("prop4: True" in self.selenium.page_source)

    def test_boolean_checkboxes_in_nested_dictionary(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop6: [True]" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox6').click()
        time.sleep(0.2)
        self.assertTrue("prop6: [False]" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox6').click()
        time.sleep(0.2)
        self.assertTrue("prop6: [True]" in self.selenium.page_source)

    def test_multi_value_checkboxes(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop5: [item1]" in self.selenium.page_source)

        # Checkbox 5
        self.selenium.find_element(By.ID, 'item1').click()
        time.sleep(0.2)
        self.assertTrue("prop5: []" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'another_item').click()
        time.sleep(0.2)
        self.assertTrue("prop5: [another_item]" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'item1').click()
        time.sleep(0.2)
        self.assertTrue("prop5: [another_itemitem1]" in self.selenium.page_source)


    def test_multi_value_checkboxes_in_nested_dictionary(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop7: [foo]" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'foo').click()
        time.sleep(0.4)
        self.assertTrue("prop7: []" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'bar').click()
        time.sleep(0.2)
        self.assertTrue("prop7: [bar]" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'foo').click()
        time.sleep(0.2)
        self.assertTrue("prop7: [barfoo]" in self.selenium.page_source)


    def test_we_require_a_value_to_be_set_for_multi_value_checkbox(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop8: []" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox8').click()
        time.sleep(0.2)
        self.assertTrue("prop8: []" in self.selenium.page_source)
        self.assertConsoleLogContains("silica:model for checkbox prop8 is set to an array but the checkbox does not have a valid value attribute")


    def test_we_require_a_valid_value_to_be_set_for_multi_value_checkbox(self):
        self.selenium.get(self.live_server_url + '/checkboxes')

        self.assertTrue("prop9: []" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'checkbox9').click()
        time.sleep(0.2)
        self.assertTrue("prop9: []" in self.selenium.page_source)
        self.assertConsoleLogContains("silica:model for checkbox prop9 is set to an array but the checkbox does not have a valid value attribute")



import time
from typing import Dict

from django.test import override_settings
from django.urls import path

from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns

TestView = create_test_view(f"""{{% silica '{__name__}.DictionaryModelComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("dict-models", TestView.as_view()),
]


class DictionaryModelComponent(Component):
    my_object: Dict = {}
    another = {
        'key1': 'value1',
    }
    nesting = {
        'key1': {
            'key2': 'value2',
        }
    }
    form = {
        'consent': False
    }

    def inline_template(self):
        return """
            <div>
                <p>
                    not permitted: '{{ my_object.doesnt_exist }}'
                    <input type="input" silica:model="my_object.doesnt_exist" id="ex1">
                </p>
                
                <p>
                    simple: '{{ another.key1 }}'
                    <input type="input" silica:model="another.key1" id="ex2">
                </p>
                
                <p>
                    2 levels: '{{ nesting.key1.key2 }}'            
                    <input type="input" silica:model="nesting.key1.key2" id="ex3">
                </p>
                
                <p>
                    checkbox consent: '{{ form.consent }}'
                    <input type="checkbox" silica:model="form.consent" id="ex4">
                </p>          
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class DictionaryModelsTestCase(SilicaBrowserTestCase):

    def test_new_property_keys_cant_be_added(self):
        self.selenium.get(self.live_server_url + '/dict-models')

        self.assertTrue("not permitted: ''" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'ex1').send_keys('n')
        time.sleep(0.5)

        self.assertTrue("not permitted: 'new'" not in self.selenium.page_source)
        self.assertConsoleLogContains('the server responded with a status of 422 (Unprocessable Entity)')

    def test_dictionaries_key_binding(self):
        self.selenium.get(self.live_server_url + '/dict-models')

        self.assertTrue("simple: 'value1'" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'ex2').send_keys('n')
        self.selenium.find_element(By.ID, 'ex2').send_keys('e')
        self.selenium.find_element(By.ID, 'ex2').send_keys('w')
        time.sleep(0.5)
        self.assertTrue("simple: 'value1new'" in self.selenium.page_source)

    def test_dictionary_keys_2_levels_deep(self):
        self.selenium.get(self.live_server_url + '/dict-models')

        self.assertTrue("2 levels: 'value2'" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'ex3').send_keys('h')
        self.selenium.find_element(By.ID, 'ex3').send_keys('e')
        self.selenium.find_element(By.ID, 'ex3').send_keys('y')
        time.sleep(0.5)
        self.assertTrue("2 levels: 'value2hey'" in self.selenium.page_source)

    def test_checkbox_model_on_nested(self):
        self.selenium.get(self.live_server_url + '/dict-models')

        self.assertTrue("checkbox consent: 'False'" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'ex4').click()
        time.sleep(0.5)
        self.assertTrue("checkbox consent: 'True'" in self.selenium.page_source)



import time

from django.test import override_settings
from django.urls import path

from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns

from selenium.webdriver.common.by import By

from silica import Component


TestView = create_test_view("""{% silica 'silica.tests.test_magic_methods.MagicMethods' %}""")


urlpatterns = silica_urlpatterns + [
    path("magic-methods", TestView.as_view()),
]

class MagicMethods(Component):
    property = "foo"

    def call_me(self):
        self.property = "boo"

    def call_me_with_arg(self, value):
        self.property = value

    def call_me_with_args(self, value1, value2):
        self.property = value1 + value2

    def inline_template(self):
        return """
            <div x-data>
                property: {{ property }}
                <button @click.prevent="$set('property', 'bar')" id="click_set">Set</button>
                <button @click.prevent="$call('call_me')" id="click_call">Call</button>
                <button @click.prevent="$call('call_me_with_arg', 'ding')" id="click_call_with_arg">Call</button>
                <button @click.prevent="$call('call_me_with_args', 'dong', 'merrily')" id="click_call_with_args">Call</button>
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class MagicMethodsBrowserTestCase(SilicaBrowserTestCase):
    def test_magic_set(self):
        self.selenium.get(self.live_server_url + "/magic-methods")

        self.assertTrue("property: foo" in self.get_page_source())

        self.selenium.find_element(By.ID, 'click_set').click()

        time.sleep(0.2)

        self.assertTrue("property: bar" in self.get_page_source())

    def test_magic_call(self):
        self.selenium.get(self.live_server_url + "/magic-methods")

        self.assertTrue("property: foo" in self.get_page_source())

        self.selenium.find_element(By.ID, 'click_call').click()

        time.sleep(0.2)

        self.assertTrue("property: boo" in self.get_page_source())

    def tmp_test_magic_call_with_arg(self):
        pass
        self.selenium.get(self.live_server_url + "/magic-methods")

        self.assertTrue("property: foo" in self.get_page_source())

        self.selenium.find_element(By.ID, 'click_call_with_arg').click()

        time.sleep(0.2)

        self.assertTrue("property: ding" in self.get_page_source())

    def tmp_test_magic_call_with_args(self):
        pass
        self.selenium.get(self.live_server_url + "/magic-methods")

        self.assertTrue("property: foo" in self.get_page_source())

        self.selenium.find_element(By.ID, 'click_call_with_arg').click()

        time.sleep(0.2)

        self.assertTrue("property: ding" in self.get_page_source())



import importlib
import json
import re
from inspect import getattr_static
from typing import Dict, Any, List
import shortuuid
import uuid

from django.apps import apps
from django.db.models import Model
from django.http import HttpRequest
from django.core.cache import cache
from django.views.generic import TemplateView
from django.template.loader import render_to_string
from django.utils.decorators import classonlymethod
from django.utils.module_loading import import_string
from silica.mixins.ValidatesProperties import ValidatesProperties

from silica.utils import pascal_to_snake
from silica.errors import SetPropertyException
from silica.SilicaTemplateResponse import SilicaTemplateResponse


class Component(TemplateView, ValidatesProperties):
    component_id: str = ""
    component_name: str = ""
    query_params: dict = {}
    has_rendered: bool = False
    js_calls: list = []
    event_calls: list = []
    lazy: bool = False

    def __init__(self, **kwargs):
        self.response_class = SilicaTemplateResponse

        self.component_id = ""
        self.component_key = ""
        self.component_name = ""
        self.lazy = False

        self.request: HttpRequest = None

        super().__init__(**kwargs)

        if "id" in kwargs:
            self.component_id = kwargs["id"]

        if "component_name" in kwargs:
            self.component_name = kwargs["component_name"]

        if "request" in kwargs:
            self.request = kwargs["request"]

        if "lazy" in kwargs:
            self.lazy = kwargs["lazy"]

        assert self.component_id != "", "Component id must be set"
        assert self.component_name != "", "Component name must be set"

        self._set_template()

    def _set_template(self):
        """Either template_name was provided, or inline_template was provided, or we'll set the template_name based 
        on the component name."""
        # check if lazy loading, if so, set template_name to placeholder
        if self.lazy:
            # todo, inline place holder -> user placeholder -> framework placeholder
            if inline_placeholder := self.inline_placeholder():
                self.template_name = f"::silica-inline::{inline_placeholder}"
                return
            # else:
            #     self.template_name = "silica/placeholder.html"
            # return

        if self.template_name:
            return

        if inline_template := self.inline_template():
            self.template_name = f"::silica-inline::{inline_template}"
            return

        self.template_name = f"silica/{self.get_template_name()}"

    # abstract
    def inline_template(self):
        pass

    # abstract
    def inline_placeholder(self):
        pass

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["component_id"] = self.component_id

        # Load local attributes into context
        local_state = self._local_state()

        # Loading state from cache into context
        cached_state = self.get_state()

        state = {**local_state, **cached_state}

        for key, value in state.items():
            context[key] = value

        # Set authenticated user
        if self.request and self.request.user.is_authenticated:
            context["user"] = self.request.user

        return context

    def _local_state(self, is_setting=False):
        state = {}
        for attr in dir(self):
            if not self._is_public(attr, is_setting):
                continue
            try:
                value = getattr(self, attr)
                if (
                        not attr.startswith("_")
                        and not callable(value)
                        and not isinstance(
                    getattr(Component, attr, None), classonlymethod
                )
                ):
                    state[attr] = value
            except AttributeError as e:
                print(f"Issue with attribute: {attr} - Error: {e}")

        return state

    def store_state(self):
        # Extracting non-private and non-callable attributes for state
        state = self._local_state(is_setting=True)

        cache.set(f"silica:component:{self.component_id}", state)

    def get_state(self):
        return cache.get(f"silica:component:{self.component_id}", {})

    def get_template_name(self):
        return pascal_to_snake(self.__class__.__name__) + ".html"

    def render(
            self, init_js=False, extra_context=None, request=None, lazy=False
    ) -> str:
        """prepare context and render component in a response"""

        if extra_context is not None:
            self.extra_context = extra_context

        if request:
            self.request = request

        context = self.get_context_data()
        context.update(
            {
                "js_calls": self.js_calls,
                "event_calls": self.event_calls,
            }
        )

        response = self.render_to_response(
            context=context,
            component=self,
            init_js=init_js,
            lazy=lazy,
        )

        if hasattr(response, "render"):
            response.render()

        rendered_component = response.content.decode("utf-8")

        return rendered_component

    @classonlymethod
    def as_view(cls, **initkwargs):
        """This is the first point into a Django view, ensure name is provided from url and initiate ID is set before the view is called."""
        if "component_id" not in initkwargs:
            initkwargs["component_id"] = str(shortuuid.uuid())

        if "component_name" not in initkwargs:
            # Convert the class name from CamelCase to snake_case
            initkwargs["component_name"] = re.sub(
                r"(?<!^)(?=[A-Z])", "-", cls.__name__
            ).lower()

        return super().as_view(**initkwargs)

    @staticmethod
    def create(
            component_id: str,
            component_name: str,
            component_key: str = "",  # will be needed sooner or later
            request: HttpRequest = None,
            kwargs: Dict[str, Any] = {},
    ) -> "Component":
        if "lazy" in kwargs:
            lazy = kwargs["lazy"]
        else:
            lazy = False

        ComponentClass = Component.get_component_class_from_name(component_name)

        component_instance = ComponentClass(
            component_id=component_id,
            component_name=component_name,
            lazy=lazy,
            request=request,
        )
        component_instance._set_state_from_cache()
        component_instance.js_calls = []
        component_instance.event_calls = []

        # Set attributes from {% silica .. %} tag
        for key, value in kwargs.items():
            if component_instance._is_public(key):
                component_instance.set_property(key, value, False)
                setattr(component_instance, key, value)

        return component_instance

    def _set_state_from_cache(self):
        """
        Load state from cache into component instance.
        """
        state = self.get_state()
        for key, value in state.items():
            setattr(self, key, value)

    def render_to_string(self):
        template_name = f"silica/{self.get_template_name()}"
        return render_to_string(template_name, self.get_context_data())

    def reset_js_calls(self):
        self.js_calls = []

    def dispatch(self, request, *args, **kwargs):
        """
        Called by the `as_view` class method when utilizing a component directly as a view.
        """

        # how to provide request in kwargs to mount()?
        self.request = request

        # if any GET variables match query_param 'at' / aliases then set the param key property, we also allow a dict
        # here that doesn't contain an 'as' key, we treat the 'param' key as the url key also
        for key, value in request.GET.items():
            for query_param in self.query_params:
                if isinstance(query_param, dict):
                    if query_param.get("as", query_param.get("param")) == key:
                        # TODO - Could this be firing twice for fresh page loads?
                        self.set_property(query_param.get("param"), value)
                        # setattr(self, query_param.get("param"), value)
                else:
                    if query_param == key:
                        # TODO - Could this be firing twice for fresh page loads?
                        self.set_property(key, value)
                        # setattr(self, key, value)

        if not self.lazy:
            self.mount()

        context = self.get_context_data()
        context.update({"js_calls": self.js_calls})
        context.update({"event_calls": self.event_calls})

        self.store_state()

        rendered_response = self.render_to_response(
            context=context,
            component=self,
            init_js=True,
        )

        return rendered_response

    def process_query_params(self) -> List[Dict[str, Any]]:
        """Determine which query params are to be shown (only show ones that are not equal to their default value)"""
        query_param_list = []

        class_obj = type(self)  # Get the component class

        for param_key in self.query_params:
            if isinstance(param_key, dict):
                # if the query param is a dict, then we'll need to handle it differently
                # we'll need to check if the param is set, and if so, set it to the value of the 'as' key
                # if not, set it to the value of the 'param' key
                param = {
                    "key": param_key.get("param"),
                    "visible": True,
                    "as": param_key.get("as", param_key.get("param")),
                }
            else:
                param = {"key": param_key, "visible": False, "as": None}

            default_value = getattr(class_obj, param["key"], None)
            current_value = getattr(self, param["key"], None)

            if current_value != default_value:
                param["visible"] = True

            query_param_list.append(param)

        return query_param_list

    def _is_public(self, name: str, for_setting: bool = False) -> bool:
        """Determines if the name should be sent in the context."""

        protected_names = (
            "render",
            "request",
            "as_view",
            "view",
            "args",
            "kwargs",
            "content_type",
            "extra_context",
            "http_method_names",
            "template_engine",
            "template_name",
            "dispatch",
            "id",
            "get",
            "get_context_data",
            "get_template_names",
            "render_to_response",
            "http_method_not_allowed",
            "options",
            "setup",
            "fill",
            "view_is_async",
            # Component methods
            "component_id",
            "component_name",
            "component_key",
            "reset",
            "mount",
            "hydrate",
            "updating",
            "update",
            "calling",
            "called",
            "complete",
            "rendered",
            "parent_rendered",
            "validate",
            "is_valid",
            "errors",
            "updated",
            "parent",
            "children",
            "call",
            "js_calls",
            "event_calls",
            "component_cache_key",
            "inline_template",
            "placeholder",
            "processed_query_params",
            "query_params",
            "validation_rules",
        )

        # Check if the attribute is a @property
        is_property = isinstance(getattr_static(self.__class__, name, None), property)

        return not (
                name.startswith("_")
                or name in protected_names
                or (for_setting and is_property)
        )

    def set_property(self, attr, value, with_hooks=True):
        """Set a property on the component."""
        segments = attr.split('.')
        base_attr = segments[0]

        if not self._is_public(base_attr, for_setting=True):
            raise SetPropertyException(f'Attribute "{base_attr}" can not be set.')

        # Single Properties on the component i.e. "day"
        if '.' not in attr:
            if hasattr(self, attr):
                # Check if the property definiton is a list and ensure the passed value is a list. This will be for 
                # query_param, multi-checkboxes or just user defined list property
                if isinstance(getattr(self, attr), list) and not isinstance(value, list):
                    value = [value]

                setattr(self, attr, value)

                if with_hooks:
                    self._invoke_hooks(attr, '', value)
            else:
                raise SetPropertyException(f'"{attr}" does not exist on component "{self.__class__.__name__}".')
            return

        # Model Properties i.e. "user.first_name" If we have validation_rules and the current segment is a Model 
        # instance, we'll need to check if the segment is a key in the validation_rules
        elif hasattr(self, 'validation_rules') and isinstance(getattr(self, base_attr), Model):
            if attr not in self.validation_rules:
                raise SetPropertyException(f'"{attr}" is not a valid key in the validation_rules.')

            setattr(getattr(self, base_attr), segments[1], value)

        else:
            # Nested Non-Model Properties i.e. "some_dict.profile.age"
            obj = self
            for i, segment in enumerate(segments[:-1]):
                # Check if the segment is looking like a list index
                if segment.isdigit() and isinstance(obj, list):
                    segment = int(segment)
                    if segment >= len(obj):
                        raise SetPropertyException(f'Index "{segment}" out of range for list.')

                    obj = obj[segment]

                else:
                    # Check for dict or attribute existence
                    if isinstance(obj, dict) and segment in obj:
                        obj = obj[segment]
                    elif hasattr(obj, segment):
                        obj = getattr(obj, segment)
                    else:
                        raise SetPropertyException(f'Dictionary key or attribute "{segment}" does not exist.')

                if not isinstance(obj, (dict, list, type(self))):
                    raise SetPropertyException('Cannot set property on non-dict/list/component property.')

            # Set the value on the final target
            final_segment = segments[-1]
            if isinstance(final_segment, str) and final_segment.isdigit() and isinstance(obj, list):
                final_segment = int(final_segment)
                if final_segment < len(obj):
                    obj[final_segment] = value
                else:
                    raise SetPropertyException(f'Index "{final_segment}" out of range for list.')

            if isinstance(obj, dict) and final_segment in obj:
                obj[final_segment] = value
            elif hasattr(obj, final_segment):
                setattr(obj, final_segment, value)
            else:
                raise SetPropertyException(f'"{final_segment}" does not exist on property "{attr}".')

            if with_hooks:
                self._invoke_hooks(base_attr, '.'.join(segments[1:]), value)

    def _invoke_hooks(self, base_attr, remaining_path, value):
        # Specific hook for the property
        updated_hook_name = f"updated_{base_attr}"
        if hasattr(self, updated_hook_name):
            if remaining_path:
                getattr(self, updated_hook_name)(remaining_path, value)
            else:
                getattr(self, updated_hook_name)(value)

        # General update hook
        if hasattr(self, "updated"):
            full_path = f"{base_attr}.{remaining_path}" if remaining_path else base_attr
            getattr(self, "updated")(full_path, value)

    def js_call(self, method, *args, **kwargs):
        self.js_calls.append(
            {
                "fn": method,
                "args": args,
                "kwargs": kwargs,
            }
        )

    # Events
    def dispatch_browser_event(self, event_name, payload: Dict = None):
        self.js_call("_silicaDispatchBrowserEvent", event_name, payload)

    def emit(self, event_name, payload: Dict = None):
        """Emit to all component present in the DOM"""
        self.event_calls.append(
            {
                "type": "emit",
                "name": event_name,
                "payload": payload,
            }
        )

    def emit_to(self, component_name, event_name, payload: Dict = None):
        """Emit to a specific component in the dom"""
        self.event_calls.append(
            {
                "type": "emit_to",
                "component_name": component_name,
                "name": event_name,
                "payload": payload,
            }
        )

    def receive_event(self, event_name, payload):
        # if we have a listener for this event, call it
        if hasattr(self, event_name) and callable(getattr(self, event_name)):
            getattr(self, event_name)(payload)

    def get_absolute_path(self):
        module_name = (
            self.__module__
        )  # gives the module name in which the class is defined
        class_name = self.__class__.__name__  # gives the name of the class
        return f"{module_name}.{class_name}"

    def get_component_path(self):
        absolute_path = self.get_absolute_path()
        if ".silica." in absolute_path:
            return absolute_path.split(".silica.", 1)[1].rsplit(".", 1)[0]
        else:
            # Handle this case accordingly.
            # You can return the original string or a placeholder/error message.
            return absolute_path

    # Hooks
    def mount(self):
        pass

    def rendered(self, html):
        """
        Hook that gets called after the component has been rendered.
        """
        pass

    def redirect(self, url):
        self.js_call("_silicaRedirect", url)

    def jsonable_data(self):
        state = self.get_context_data()

        for key, value in state.copy().items():
            if not self._is_jsonable(value):
                del state[key]

        return state

    def _serialize_property_value(self, value):
        if isinstance(value, uuid.UUID):
            return str(value)
        elif self._is_jsonable(value):
            return value
        else:
            return None

    def format_data(self):
        state = self.get_context_data()
        formatted = {}

        for key, value in state.copy().items():
            if self._is_jsonable(value):
                formatted[key] = value
            # check for uuid
            elif isinstance(value, uuid.UUID):
                formatted[key] = str(value)
            elif isinstance(value, Model):
                # We'll only pass back model keys that have been specified in the component's validation_rules, 
                # consider that the keys in validation_rules will be in format of "{model_instance}.{field}" and 
                # we'll want to provide them as a json object
                for rule_key in self.validation_rules.keys():
                    if rule_key.startswith(f"{key}."):
                        formatted[key] = {
                            **formatted.get(key, {}),
                            rule_key.split(".")[1]: self._serialize_property_value(
                                getattr(value, rule_key.split(".")[1])),
                        }

        return formatted

    def create_snapshot(self):
        snapshot = {
            'data': self.format_data(),
            'js_calls': self.js_calls,
            'event_calls': self.event_calls,
            'query_params': self.process_query_params(),
        }
        return snapshot

    def _is_jsonable(self, x):
        if isinstance(x, Model):
            return False
        try:
            json.dumps(x)
            return True
        except (TypeError, OverflowError):
            return False

    @staticmethod
    def get_component_class_from_name(component_name: str):
        for app in apps.get_app_configs():
            # Attempt to load the 'silica' module from each app
            try:
                silica_module = importlib.import_module(
                    f"{app.name}.silica.{component_name}"
                )

                importlib.reload(silica_module)

                # class_name will always be the last portion of the path, the last part of the string after the last 
                # dot, dots may not exist
                class_name = component_name.split(".")[-1]

                # If we don't encounter an ImportError, try to get the component
                component_class = getattr(silica_module, class_name, None)
                if component_class:
                    # If found, use it (or return, or whatever your logic is)
                    return component_class
            except ImportError:
                continue

        # finally, try and import the component_name itself if it's a direct path to a component
        try:
            component_class = import_string(component_name)
            if component_class:
                # If found, use it (or return, or whatever your logic is)
                return component_class
        except ImportError:
            pass

        # If the component wasn't found in any app's 'silica' module, raise an error
        raise ImportError(
            f"Component {component_name} not found in any app's 'silica' module"
        )

    def __repr__(self):
        return f"Component(name='{self.component_name}' id='{self.component_id}')"



import time
from typing import List, Dict

from django.test import override_settings
from django.urls import path

from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.Parent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("nestable", TestView.as_view()),
]

class Parent(Component):
    dummy = ''

    def inline_template(self):
        return """
            <div>
                Parent: "{{ dummy }}"
                <input type="test" silica:model="dummy">
                {% silica '"""+__name__+""".Child' %}             
            </div>
        """

class Child(Component):
    dummy = ''

    def dummy_updated(self, value):
        self.js_call('alert', 'hi')

    def inline_template(self):
        return """
            <div>
                Child: "{{ dummy }}"
                <input type="test" silica:model="dummy" id="child_input">             
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class NestedTestCase(SilicaBrowserTestCase):

    def test_child_input_doesnt_trigger_parent(self):

        self.selenium.get(self.live_server_url + '/nestable')

        self.assertTrue("child_input" in self.selenium.page_source)

        self.selenium.find_element(By.ID, 'child_input').send_keys('h')
        time.sleep(0.2)

        self.selenium.save_screenshot('screenshot.png')
        # screenshot
        self.assertTrue('Parent: ""' in self.selenium.page_source)
        self.assertTrue('Child: "h"' in self.selenium.page_source)




import time
from unittest import skip

from django.test import override_settings
from django.urls import path


from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


class BrowserEventDispatcher(Component):
    def mount(self):
        self.dispatch_browser_event("test", {"value": 123})

    def inline_template(self):
        return """
            <div>
                <div x-data="{
                    log() {
                        console.log(`event called with payload: ${$event.detail.value}`)                    
                    }
                }">
                    <div @test.window="log"></div>
                </div>
            </div>
        """


TestView = create_test_view(f"""
    {{% silica '{__name__}.BrowserEventDispatcher' %}}
    """)


urlpatterns = silica_urlpatterns + [
    path("dispatch-browser-events/", TestView.as_view()),
]


@override_settings(ROOT_URLCONF=__name__)
class DispatchBrowserEventTestCase(SilicaBrowserTestCase):

    @skip("This test is not working")
    def test_browser_event_received(self):
        self.selenium.get(self.live_server_url + '/dispatch-browser-events')
        time.sleep(0.4)
        self.assertConsoleLogContains("event called with payload: 123")



from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class Redirect(Component):
    def redirect_me(self, path):
        self.redirect(path)

    def inline_template(self):
        return """
            <div class="p-5 border m-5">
                <p>I will be redirected!</p>
            </div>
        """


class RedirectTest(SilicaTestCase):
    def test_redirect_from_method(self):
        (
            SilicaTest(component=Redirect)
            .call("redirect_me", "/redirected")
            .assertJsCalled("_silicaRedirect", "/redirected")
        )



from time import sleep

from django.test import override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.MagicProxy' %}}""")

urlpatterns = silica_urlpatterns + [
    path("proxy", TestView.as_view()),
]


class MagicProxy(Component):
    prop = "hello"
    items = [1,2,3,4]

    def updated_prop(self, value):
        pass

    def inline_template(self):
        return """
            <div x-data>
                Getter: <span x-text="`hello ${$silica.prop}`"></span>
                <button id="update_prop" @click="$silica.prop = 'world'">Set</button>
                prop: {{ prop }}
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class MagicProxyTest(SilicaBrowserTestCase):

    def test_magice_silica_proxy(self):
        self.selenium.get(self.live_server_url + '/proxy')

        self.assertHtmlContains('hello hello')
        self.findElementById('update_prop').click()
        sleep(0.3)
        self.assertHtmlContains("prop: world")






from silica import Component


class ShortName(Component):
    def inline_template(self):
        return """
            <div>
                I'm component called with short name!
            </div>
        """



from django.test import TestCase
from django.core.management import call_command
from django.conf import settings
from unittest.mock import patch
from pathlib import Path

from silica.utils import pascal_to_snake

class SilicaRmCommandTestCase(TestCase):
    def setUp(self):
        self.files = []
        self.base_dir = Path(getattr(settings, 'BASE_DIR', None))
        self.main_app_name = 'example_project'

    def tearDown(self):
        for file in self.files:
            if file.exists():
                file.unlink()

    def create_temp_component(self, component_name, subfolders=[]):
        snake_component_name = pascal_to_snake(component_name)
        component_path = self.base_dir / self.main_app_name / 'silica' / Path(*subfolders) / f"{component_name}.py"
        template_path = self.base_dir / self.main_app_name / 'templates' / 'silica' / Path(*subfolders) / f"{snake_component_name}.html"
        component_path.parent.mkdir(parents=True, exist_ok=True)
        template_path.parent.mkdir(parents=True, exist_ok=True)
        component_path.touch()
        template_path.touch()
        self.files.extend([component_path, template_path])
        return component_path, template_path

    def test_remove_component_success(self):
        component_name = "TestComponent"
        component_path, template_path = self.create_temp_component(component_name)

        with patch('builtins.input', return_value='y'):
            call_command('silica', 'rm', component_name, "--silent")

        self.assertFalse(component_path.exists(), "Component file was not removed.")
        self.assertFalse(template_path.exists(), "Template file was not removed.")

    def test_remove_component_skip_confirmation(self):
        component_name = "TestComponent"
        component_path, template_path = self.create_temp_component(component_name)

        call_command('silica', 'rm', component_name, '-y', "--silent")

        self.assertFalse(component_path.exists(), "Component file was not removed.")
        self.assertFalse(template_path.exists(), "Template file was not removed.")

    def test_remove_nonexistent_component(self):
        component_name = "NonExistentComponent"
        # Convert component name to snake_case for the template
        snake_component_name = pascal_to_snake(component_name)

        # Construct paths for the component and template
        component_path = self.base_dir / self.main_app_name / 'silica' / f"{component_name}.py"
        template_path = self.base_dir / self.main_app_name / 'templates' / 'silica' / f"{snake_component_name}.html"

        # Ensure the files do not exist before command execution
        self.assertFalse(component_path.exists(), "Non-existent component file found.")
        self.assertFalse(template_path.exists(), "Non-existent template file found.")

        # Execute the command
        call_command('silica', 'rm', component_name, "--silent")

        # Assert that the files still do not exist after command execution
        self.assertFalse(component_path.exists(), "Non-existent component file was mistakenly created.")
        self.assertFalse(template_path.exists(), "Non-existent template file was mistakenly created.")

    # Additional tests for handling subfolders, invalid component names, etc., can be added as needed.



import time

from django.test import override_settings
from django.urls import path

from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.urls import urlpatterns as silica_urlpatterns


## The test components
class ScalarAssignment(Component):
    test_value: str = None

    def inline_template(self):
        # language=HTML
        return """
            <div>
                <button silica:click.prevent="test_value = 1" id="button">Set</button>
                test value: {{ test_value }}
            </div>
        """


## Urls for the test
urlpatterns = silica_urlpatterns + [
    path("silica/tests/scalar-assignment", ScalarAssignment.as_view()),
]


## The test case
@override_settings(ROOT_URLCONF=__name__)
class PropertyExpressionAssignmentTestCase(SilicaBrowserTestCase):

    def test_scalar_assignment_expression(self):
        self.selenium.get(self.live_server_url + '/silica/tests/scalar-assignment')

        self.selenium.find_element(By.ID, 'button').click()

        # Give time for the requests to complete
        time.sleep(1)

        # Assert that test_value is set to 1 in the dom
        self.assertIn('test_value = 1', self.selenium.page_source)



from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class InlineTemplate(Component):
    msg = "Hello World!"

    def inline_template(self):
        return """<div>{{ msg }}</div>"""


class LifecycleTestCase(SilicaTestCase):
    def test_inline_template_is_rendered(self):
        SilicaTest(component=InlineTemplate).assertSee("Hello World!")

    def test_subsequent_requests_render_from_inline_template(self):
        (
            SilicaTest(component=InlineTemplate)
            .assertSee("Hello World!")
            .set("msg", "Goodbye World!")
            .assertSee("Goodbye World!")
        )



from json import loads

from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from silica import Component
from silica.errors import SetPropertyException


@require_POST
@csrf_exempt
def message(request, *args, **kwargs):
    """endpoint that processes dynamic updates to components"""

    request_body = loads(request.body)

    component_id = request_body.get("id")
    component_name = request_body.get("name")

    actions = request_body.get("actions")

    assert component_id, "Component id must be set"
    assert component_name, "Component name must be set"

    component_instance = Component.create(
        component_id=component_id, 
        component_name=component_name, 
        request=request
    )

    # if actions is a List, for each action, execute
    if actions:
        for action in actions:
            if action["type"] == "set_property":
                name = action.get("name")
                value = action.get("value")

                try:
                    component_instance.set_property(name, value)
                except SetPropertyException as e:
                    return JsonResponse(
                        {
                            "error": f"Component {component_name}: {e}"
                        },
                        status=422,
                    )

            elif action["type"] == "call_method":
                method = getattr(component_instance, action["method_name"])
                method(*action.get("args", []), **action.get("kwargs", {}))
            elif action["type"] == "activate_lazy":
                component_instance.mount()
            elif action["type"] == "event":
                component_instance.receive_event(action["event_name"], action["payload"])

    component_instance.store_state()

    html = component_instance.render()

    snapshot = component_instance.create_snapshot()

    return JsonResponse(
        {
            "html": html,
            "id": component_id,
            "snapshot": snapshot,
        }
    )




from django.urls import path
from django.views.generic import TemplateView


class ComponentTagTest(TemplateView):
    template_name = "component_tag_test.html"


class ComponentSubfolderTest(TemplateView):
    template_name = "component_subfolder_test.html"

urlpatterns = [
    path("silica/tests/component-tag-test", ComponentTagTest.as_view()),
    path("silica/tests/component-subfolder-test", ComponentSubfolderTest.as_view()),
]



from typing import TypeVar


class Paginator:
    """A simple paginator class that returns page details for a given 1-based page number."""

    def __init__(self, total_count, per_page=10, page_fn=None):
        self.total_count = total_count
        self.per_page = per_page
        self.num_pages = self._get_num_pages()
        self.page_fn = page_fn

    def _get_num_pages(self):
        """Returns the total number of pages."""
        if self.total_count % self.per_page == 0:
            return self.total_count // self.per_page
        return self.total_count // self.per_page + 1

    def validate_number(self, number):
        """Validates the given 1-based page number."""
        try:
            number = int(number)
        except (TypeError, ValueError):
            raise ValueError(f"Invalid page number: {number}")
        if number < 1 or number > self.num_pages:
            raise ValueError(f"Page number out of range: {number}")
        return number

    def page(self, number):
        """Returns page details for the given 1-based page number."""
        number = self.validate_number(number)
        offset = (number - 1) * self.per_page
        limit = self.per_page

        # # If a page function is provided, use it to fetch data.
        # if self.page_fn:
        #     object_list = self.page_fn(offset, limit)
        # else:
        #     object_list = range(offset, limit)  # Default behavior for demonstration

        return {
            'number': number,
            'offset': offset,
            'limit': limit,
            'has_previous': number > 1,
            'has_next': number < self.num_pages,
            'num_pages': self.num_pages,
            'next_page_number': number + 1 if number < self.num_pages else None,
            'previous_page_number': number - 1 if number > 1 else None,
        }

    __repr__ = __str__ = lambda self: f"<Paginator with {self.num_pages} pages>"


# create a type from Paginator class page() method:
PaginatorPage = TypeVar("PaginatorPage", bound=Paginator.page)



from datetime import datetime

from django.core.exceptions import ValidationError


class ValidatesProperties:
    validation_rules = {}

    def validate(self, data):
        """
        Validate the component's properties.
        """
        validation_rules = getattr(self, "validation_rules", None)
        if not validation_rules:
            return

        errors = {}

        for field, rules in validation_rules.items():
            value = data.get(field)
            for rule in rules:
                rule_parts = rule.split(":")
                rule_name = rule_parts[0]
                rule_args = rule_parts[1:] if len(rule_parts) > 1 else []

                method_name = f"validate_{rule_name}"
                method = getattr(self, method_name, None)

                if method:
                    error = method(value, *rule_args)
                    if error:
                        errors.setdefault(field, []).append(error)

        if errors:
            raise ValidationError(errors)

        return errors

    # Validators
    def validate_required(self, value):
        if value is None or (isinstance(value, str) and not value.strip()):
            return "This field is required."

    def validate_email(self, value):
        pattern = r"[^@]+@[^@]+\.[^@]+"
        if not re.match(pattern, value):
            return "Invalid email format."

    def validate_min_length(self, value, length):
        if len(value) < length:
            return f"This field must be at least {length} characters."

    def validate_max_length(self, value, length):
        if len(value) > length:
            return f"This field must not exceed {length} characters."

    def validate_date_equals(self, value, compare_date):
        try:
            date_value = datetime.strptime(value, "%Y-%m-%d")
            if date_value.date() != compare_date.date():
                return "The date does not match the given date."
            return None
        except ValueError:
            return "The value is not a valid date."

    def validate_exists(self, value, data_list):
        # Simplified exists check against a list. In Laravel, it checks against a DB.
        if value not in data_list:
            return "The selected value is invalid."
        return None

    def validate_file(self, file_obj):
        if not hasattr(file_obj, "read"):
            return "The value is not a valid file."
        return None

    def validate_integer(self, value):
        if not isinstance(value, int):
            return "The value is not an integer."
        return None

    def validate_nullable(self, value):
        # If value is None or an empty string, validation passes. Else, it fails.
        if value in [None, ""]:
            return None
        return "The value is not nullable."

    def validate_alpha_dash(self, value):
        if not re.match("^[a-zA-Z0-9_-]*$", value):
            return (
                "The value may only contain letters, numbers, dashes, and underscores."
            )
        return None

    def validate_alpha_num(self, value):
        if not value.isalnum():
            return "The value may only contain letters and numbers."
        return None

    def validate_array(self, value):
        if not isinstance(value, list):
            return "The value must be an array."
        return None

    def validate_url(self, value):
        # Placeholder: This is a complex one in Laravel, requiring DNS checks. Here's a simplified version:
        import validators

        if not validators.url(value):
            return "The URL is not valid."
        return None

    def validate_max(self, value, max_val):
        if len(value) > max_val:
            return f"The value must not be greater than {max_val} characters."
        return None

    def validate_min(self, value, min_val):
        if len(value) < min_val:
            return f"The value must be at least {min_val} characters."
        return None

    def validate_not_in(self, value, *args):
        if value in args:
            return "The selected value is invalid."
        return None

    def validate_regex(self, value, pattern):
        if not re.match(pattern, value):
            return "The value format is invalid."
        return None

    def validate_starts_with(self, value, *args):
        if not value.startswith(args):
            return "The value must start with one of the following: " + ", ".join(args)
        return None

    def validate_numeric(self, value):
        if not isinstance(value, (int, float)):
            return "The value must be numeric."
        return None

    def validate_boolean(self, value):
        if not isinstance(value, bool):
            return "The value must be boolean."
        return None

    def validate_confirmed(self, value, confirmation_field):
        if value != self.data.get(confirmation_field):
            return "The confirmation does not match."
        return None

    def validate_date(self, value, format="%Y-%m-%d"):
        try:
            datetime.strptime(value, format)
            return None
        except ValueError:
            return "The value is not a valid date."

    def validate_before(self, value, date_string):
        input_date = datetime.strptime(value, "%Y-%m-%d")
        comparison_date = datetime.strptime(date_string, "%Y-%m-%d")
        if input_date >= comparison_date:
            return f"The date must be before {date_string}."
        return None

    def validate_after(self, value, date_string):
        input_date = datetime.strptime(value, "%Y-%m-%d")
        comparison_date = datetime.strptime(date_string, "%Y-%m-%d")
        if input_date <= comparison_date:
            return f"The date must be after {date_string}."
        return None

    def validate_between(self, value, min_val, max_val):
        if not min_val <= len(value) <= max_val:
            return f"The value length must be between {min_val} and {max_val}."
        return None

    def validate_different(self, value, other_field):
        if value == self.data.get(other_field):
            return f"The value must be different from {other_field}."
        return None

    def validate_alpha(self, value):
        if not value.isalpha():
            return "The value must be alphabetic."
        return None



from django.template import TemplateDoesNotExist
from django.template import Template
from django.template.loaders.base import Loader


class Loader(Loader):
    def get_template_sources(self, template_name):
        return [template_name]

    def get_contents(self, origin):
        return origin

    def get_template(self, template_name, skip=None, dirs=None):
        """We identify if this is an silica inline template with the '::silica-inline::' prefix"""
        if "::silica-inline::" not in template_name:
            raise TemplateDoesNotExist("Template %s not found" % template_name)

        inline_template = template_name.replace("::silica-inline::", "")

        return Template(inline_template)

        # TO DEPRECATE
        # is_admin = False
        # for template_source in self.get_template_sources(template_name):
        #     if "admin" in template_source:
        #         is_admin = True
        #
        # if is_admin:
        #     raise TemplateDoesNotExist("Template %s not found" % template_name)
        # else:
        #     return Template(template_name, engine=self.engine)



from django.test import override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


class Component(Component):
    count = 1
    change_prop = False

    def mount(self):
        if self.change_prop:
            self.count += 1

    def inline_template(self):
        return """
            <div>
                count:{{ count }} 
            </div>
        """


Page1View = create_test_view("""    
    <h1>Page 1</h1>
    <a href="page_2" id="visit_page_2">page 2</a>
    {% silica 'silica.tests.test_class_property_modification.Component' change_prop='True' %}
""")


Page2View = create_test_view("""   
    <h1>Page 2</h1>
    {% silica 'silica.tests.test_class_property_modification.Component' change_prop='False' %}
""")


urlpatterns = silica_urlpatterns + [
    path("page_1", Page1View.as_view()),
    path("page_2", Page2View.as_view())
]


@override_settings(ROOT_URLCONF=__name__)
class ClassPropertyModTestCase(SilicaBrowserTestCase):
    def test_modifying_class_property_doesnt_affect_future_instances(self):

        self.selenium.get(self.live_server_url + '/page_1')
        self.assertTrue('count:2' in self.selenium.page_source)

        self.selenium.get(self.live_server_url + '/page_2')
        self.assertTrue('count:1' in self.selenium.page_source)


class SilicaCacheError(Exception):
    pass


class SetPropertyException(Exception):
    pass


class ComponentNotValid(Exception):
    pass


class SilicaSetPropertyException(Exception):
    pass


class SingleRootNodeError(Exception):
    pass



import json
from urllib.parse import unquote
from typing import Type, List

from bs4 import BeautifulSoup

from django.core.handlers.wsgi import WSGIRequest
from django.test import TestCase, RequestFactory, override_settings
from django.contrib.auth.models import AnonymousUser, User

from silica import Component
from silica.views import message
from silica.tests.urls import urlpatterns as silica_test_url_patterns
from silica.urls import urlpatterns as silica_urlpatterns

from silica.tests.utils import data_get

urlpatterns = silica_urlpatterns + silica_test_url_patterns


@override_settings(ROOT_URLCONF=__name__)
class SilicaTestCase(TestCase):
    pass


@override_settings(ROOT_URLCONF=__name__)
class SilicaTest(TestCase):
    received_html = ""
    received_data = {}
    received_js_calls = []
    received_event_calls = []
    _user: User = None

    def __init__(
            self,
            component: Type[Component],
            request: Type[WSGIRequest] = None,
            user: User = None,
            **kwargs
    ):
        super().__init__()

        if request is None:
            request = RequestFactory().get("/", **kwargs)
            request.user = AnonymousUser()
            
        if user is not None:
            request.user = user
            self._user = user

        response = component.as_view(**kwargs)(request)

        self.component = response.component
        self.received_html = response.render().content.decode("utf-8")

        # use beautiful soup to parse the html
        soup = BeautifulSoup(self.received_html, "html.parser")

        # extract the attribute: silica:initial-data from the root node
        root_node = soup.contents[0]
        initial_data = root_node.attrs.get("silica:initial-data", None)

        # uri decode the initial data
        if initial_data:
            initial_data_raw = unquote(initial_data)
            initial_data_parsed = json.loads(initial_data_raw)

            self.received_event_calls = initial_data_parsed.pop('event_calls')
            self.received_js_calls = initial_data_parsed.pop('js_calls')
            self.received_data = initial_data_parsed

    def call(self, method_name, *args, **kwargs):
        self._send_message(
            [{"type": "call_method", "method_name": method_name, "args": args}]
        )

        return self

    def set(self, prop, value):
        self._send_message([{"type": "set_property", "name": prop, "value": value}])

        return self

    def assertSet(self, prop, value):
        """
        Assert that a property is set to a specific value or satisfies a given condition.

        Parameters:
        prop (str): The property to check, using dot notation for nested properties.
        value (Union[Any, Callable[[dict], bool]]): 
            Either the expected value, or a lambda function that takes the received_data
            as an argument and returns a boolean.

        Returns:
        self: Returns self for method chaining.

        Raises:
        AssertionError: If the assertion fails.
        """

        if callable(value):
            # If value is a callable (lambda), pass the received_data to it
            condition_met = value(self.received_data)
            self.assertTrue(condition_met, f"Condition for property '{prop}' was not met")
        else:
            # If value is not a callable, compare directly
            actual_value = data_get(self.component, prop, None)
            self.assertEqual(actual_value, value, f"Property '{prop}' is not set to '{value}'")

        return self

    def assertSee(self, text):
        assert text in self.received_html, f'"{text}" is not seen in the rendered html'
        return self

    def assertDontSee(self, text):
        assert text not in self.received_html, f'"{text}" is seen in the rendered html'
        return self

    def assertNone(self, prop):
        self.assertIsNone(getattr(self.component, prop, None))
        return self

    def assertJsCalled(self, fn_name, *args):
        self.assertTrue(isinstance(self.received_js_calls, list))

        if args:
            def condition(call):
                return call["fn"] == fn_name and call["args"] == list(args)
        else:
            def condition(call):
                return call["fn"] == fn_name

        self.assertTrue(any(condition(call) for call in self.received_js_calls))

        return self

    def assertEventEmitted(self, event_name, event_payload=None):
        self.assertTrue(
            type(self.received_event_calls) == list,
            'event_calls not seen in the data'
        )

        # assert that a dict in list contains a key 'name' matching the event_name, optionally check if the event_payload matches too
        self.assertTrue(
            any(call['name'] == event_name and (event_payload is None or call['payload'] == event_payload) for call in
                self.received_event_calls),
            f'Event {event_name} not seen in the data'
        )

        return self

    def assertEventNotEmitted(self, event_name):
        self.assertTrue(
            type(self.received_event_calls) == list,
            'event_calls not seen in the data'
        )

        self.assertFalse(
            any(call['name'] == event_name for call in self.received_event_calls),
            f'Event {event_name} seen in the data'
        )

        return self

    def emit(self, event_name, payload):
        self._send_message([
            {
                "type": "event",
                "event_name": event_name,
                "payload": payload
            }
        ])

        return self

    def emit_to(self, event_name, payload):
        self._send_message([
            {
                "type": "event",
                "event_name": event_name,
                "payload": payload
            }
        ])

        return self

    def _send_message(self, actions: List):
        # prepare request
        data = json.dumps(
            {
                "id": self.component.component_id,
                "name": self.component.get_component_path(),
                "actions": actions,
            }
        )

        factory = RequestFactory()
        request = factory.post(
            "/silica/message", data, content_type="application/json"
        )

        if self._user is not None:
            request.user = self._user
        else:
            request.user = AnonymousUser()

        json_response = message(request)
        response_data = json.loads(json_response.content.decode("utf-8"))

        if 'error' in response_data:
            raise Exception(response_data['error'])

        self.received_html = response_data["html"]

        snapshot = response_data["snapshot"]
        self.received_data = snapshot["data"]
        self.received_js_calls = snapshot["js_calls"]
        self.received_event_calls = snapshot["event_calls"]

        # rehydrate the test component
        self.component._set_state_from_cache()



from typing import Dict

import shortuuid

from silica.utils import kwarg_as_dict
from silica.errors import ComponentNotValid
from silica import Component
from django import template

register = template.Library()


def silica(parser, token):
    parts = token.split_contents()

    if len(parts) < 2:
        raise template.TemplateSyntaxError(
            "%r tag requires at least a single argument" % token.contents.split()[0]
        )

    component_name = parser.compile_filter(parts[1])

    kwargs = {}
    for arg in parts[2:]:
        try:
            kwarg = kwarg_as_dict(arg)
            kwargs.update(kwarg)
        except ValueError:
            pass

    return SilicaNode(component_name=component_name, kwargs=kwargs)


register.tag("silica", silica)


class SilicaNode(template.Node):
    def __init__(self, component_name, kwargs: Dict = {}):
        self.component_name = component_name
        self.kwargs = kwargs
        self.component_key = ""
        self.parent = None

    def render(self, context, **kwargs):
        request = context.get("request", None)

        resolved_kwargs = {}
        for key, value in self.kwargs.items():
            try:
                resolved_value = template.Variable(value).resolve(context)
                resolved_kwargs.update({key: resolved_value})
            except template.VariableDoesNotExist:
                resolved_kwargs.update({key: value})

        if "key" in resolved_kwargs:
            component_key = resolved_kwargs.pop("key")
        else:
            component_key = None

        if "lazy" in resolved_kwargs:
            lazy = resolved_kwargs.pop("lazy")
        else:
            lazy = False

        try:
            component_name = self.component_name.resolve(context)
        except AttributeError:
            raise ComponentNotValid(
                f"Component template is not valid: {self.component_name}."
            )

        resolved_kwargs.update({"lazy": lazy})

        component_id = shortuuid.uuid()

        self.view = Component.create(
            component_id=component_id,
            component_name=component_name,
            component_key=component_key,
            kwargs=resolved_kwargs,
            request=request,
        )

        if request:

            # Oddly, django's request.GET will only either return dict (.items(), .dict()) where it will cut short list/multiples of one key
            # or .lists() where it will force all items as a list, even single values, but we do get the multiples.
            processed_params = {}

            for key, values in request.GET.lists():
                # Store the value directly if there's only one, or as a list if there are multiple
                processed_params[key] = values if len(values) > 1 else values[0]

            for key, value in processed_params.items():
                for query_param in self.view.query_params:
                    if isinstance(query_param, dict):
                        if query_param.get("as", query_param.get("param")) == key:
                            # TODO, should probably not allow dictionary key path setting, just limit to base attribute
                            self.view.set_property(query_param.get("param"), value, False)
                    else:
                        if query_param == key:
                            # TODO, should probably not allow dictionary key path setting, just limit to base attribute
                            self.view.set_property(key, value, False)

            self.request = request

        if not lazy:
            self.view.mount()

        rendered_component = self.view.render(init_js=True, request=request)

        return rendered_component



from django.test import Client, override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaTestCase import SilicaTest, SilicaTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


class TagProps(Component):
    bool_val: bool = False
    no_value = None
    # int_val: int = 0
    # float_val: float = 0.0
    # str_val: str = ""
    # list_val: list = []
    # dict_val: dict = {}
    # none_val: None = None

    def inline_template(self):
        return """
            <div>   
                bool_val: {{ bool_val }}<br>
                no_value: {{ no_value }}<br>
            </div>
        """


TestView = create_test_view("""    
    {% silica 'silica.tests.test_tag_props.TagProps' bool_val="True" %}
""")


urlpatterns = silica_urlpatterns + [
    path("tag-props", TestView.as_view()),
]


@override_settings(ROOT_URLCONF=__name__)
class TestTagProps(SilicaTestCase):
    def test_props_can_be_set_programmatically(self):
        (
            SilicaTest(component=TagProps, bool_val=None)
            .assertSet("bool_val", None)
        )

    def test_props_can_be_set_via_tag(self):
        client = Client()
        response = client.get("/tag-props")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "bool_val: True")


import time

from django.test import override_settings
from django.urls import path
from selenium.webdriver.common.by import By

from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica import Component
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns

TestView = create_test_view("""{% silica 'silica.tests.test_script_tags_hydrated.TestComponent' %}""")


urlpatterns = silica_urlpatterns + [
    path("hydrate-scripts", TestView.as_view())
]


class TestComponent(Component):
    show_content = False

    def inline_template(self):
        return """
            <div>
                {% if show_content %}
                    <script>
                        console.log('yoo')
                    </script>
                {% endif %}
                <button silica:click.prevent="show_content = 1" id="button">Show content</button>
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class ScriptHydrationTestCase(SilicaBrowserTestCase):
    def test_script_tag_is_hydrated(self):
        self.selenium.get(self.live_server_url + "/hydrate-scripts")

        self.selenium.find_element(By.ID, 'button').click()
        time.sleep(0.2)

        self.assertConsoleLogContains("yoo")







from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class JsCalls(Component):
    def set_js_call(self):
        self.js_call("alert", "hi")

    def inline_template(self):
        return "<div>testing</div>"


class TestJsCalls(SilicaTestCase):
    def test_js_is_called(self):
        (
            SilicaTest(component=JsCalls)
            .call("set_js_call")
            .assertJsCalled("alert", "hi")
        )



from io import StringIO
import os
from silica.utils import pascal_to_snake
from django.core.management.base import BaseCommand, CommandError

def snake_to_camel(snake_str):
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


class Command(BaseCommand):
    help = 'Move a Silica component'
    example = 'silica_mv existing_component_name new_component_name'

    def add_arguments(self, parser):
        parser.add_argument('existing_component_name', type=str)
        parser.add_argument('new_component_name', type=str)
        parser.add_argument('--app', type=str, help='Specify the Django app name')
        parser.add_argument('--silent', action='store_true', help='Run the command in silent mode without output')

    def handle(self, *args, **options):
        existing_component_name: str = options['existing_component_name']
        new_component_name: str = options['new_component_name']
        app_name = options['app']
        existing_subfolders: list[str] = []
        new_subfolders: list[str] = []
        silent = options['silent']


        if silent:
            self.stdout = StringIO()
    
        if '.' in existing_component_name:
            *existing_subfolders, existing_component_name = existing_component_name.split(".")
        if '.' in new_component_name:
            *new_subfolders, new_component_name = new_component_name.split(".")
        
        existing_subfolders = [subfolder.lower() for subfolder in existing_subfolders]
        new_subfolders = [subfolder.lower() for subfolder in new_subfolders]
        existing_component_name = existing_component_name.replace('-', '_')
        new_component_name = new_component_name.replace('-', '_')

        if "_" in existing_component_name:
            existing_component_name = snake_to_camel(existing_component_name)
        else:
            # Capitalize the first letter
            existing_component_name = existing_component_name[0].upper() + existing_component_name[1:]

        if "_" in new_component_name:
            new_component_name = snake_to_camel(new_component_name)
        else:
            # Capitalize the first letter
            new_component_name = new_component_name[0].upper() + new_component_name[1:]

        # Convert component name to snake_case for the template
        existing_snake_component_name = pascal_to_snake(existing_component_name)
        new_snake_component_name = pascal_to_snake(new_component_name)

        if app_name:
            main_app_name = app_name
        else:
            settings_module = os.environ.get('DJANGO_SETTINGS_MODULE')
            if not settings_module:
                raise CommandError("DJANGO_SETTINGS_MODULE environment variable not set.")
            main_app_name = settings_module.split('.')[0]

        # Get the BASE_DIR from Django settings
        from django.conf import settings
        base_dir = getattr(settings, 'BASE_DIR')
        if not base_dir:
            raise CommandError("Unable to determine the BASE_DIR from Django settings.")


        # component
        existing_component_path = os.path.join(base_dir, main_app_name, 'silica', *existing_subfolders, f"{existing_component_name}.py")

        if not os.path.exists(existing_component_path):
            raise CommandError(f"Component {existing_component_name} does not exist.")
        
        new_component_path = os.path.join(base_dir, main_app_name, 'silica', *new_subfolders, f"{new_component_name}.py")

        if os.path.exists(new_component_path):
            raise CommandError(f"Component {new_component_name} already exists.")
        

        # template
        existing_template_path = os.path.join(base_dir, main_app_name, 'templates', 'silica', *existing_subfolders, f"{existing_snake_component_name}.html")

        if not os.path.exists(existing_template_path):
            raise CommandError(f"Template {existing_snake_component_name}.html does not exist.")
        
        new_template_path = os.path.join(base_dir, main_app_name, 'templates', 'silica', *new_subfolders, f"{new_snake_component_name}.html")

        if os.path.exists(new_template_path):
            raise CommandError(f"Template {new_snake_component_name}.html already exists.")
        

        # Move component
        os.rename(existing_template_path, new_template_path)
        os.rename(existing_component_path, new_component_path)

        self.stdout.write(self.style.SUCCESS(f"Component {existing_component_name} moved to {new_component_name}"))
        self.stdout.write(self.style.SUCCESS(f"Template {existing_snake_component_name}.html moved to {new_snake_component_name}.html"))



from io import StringIO
import os
from silica.utils import pascal_to_snake
from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string


def snake_to_camel(snake_str):
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


class Command(BaseCommand):
    concept = """rm [component] - remove the component file and template path, with a prompt "Are you sure you want to remove this component?"""
    help = "remove a Silica component"
    example = "silica_rm component_name"

    def add_arguments(self, parser):
        parser.add_argument('component_name', type=str)
        parser.add_argument('--app', type=str, help='Specify the Django app name')
        parser.add_argument('-y', action='store_true', help='Do not prompt for confirmation')
        parser.add_argument('--silent', action='store_true', help='Run the command in silent mode without output')

    def handle(self, *args, **options):
        component_name: str = options['component_name']
        app_name = options['app']
        skip_confirmation = options['y']
        subfolders: list[str] = []
        silent = options['silent']


        if silent:
            self.stdout = StringIO()
    
        if '.' in component_name:
            *subfolders, component_name = component_name.split(".")

        
        subfolders = [subfolder.lower() for subfolder in subfolders]
        component_name = component_name.replace('-', '_')

        if "_" in component_name:
            component_name = snake_to_camel(component_name)
        else:
            # Capitalize the first letter
            component_name = component_name[0].upper() + component_name[1:]

        # Convert component name to snake_case for the template
        existing_snake_component_name = pascal_to_snake(component_name)

        if app_name:
            main_app_name = app_name
        else:
            settings_module = os.environ.get('DJANGO_SETTINGS_MODULE')
            if not settings_module:
                raise CommandError("DJANGO_SETTINGS_MODULE environment variable not set.")
            main_app_name = settings_module.split('.')[0]

        # Get the BASE_DIR from Django settings
        from django.conf import settings
        base_dir = getattr(settings, 'BASE_DIR')
        if not base_dir:
            raise CommandError("Unable to determine the BASE_DIR from Django settings.")


        # component
        component_path = os.path.join(base_dir, main_app_name, 'silica', *subfolders, f"{component_name}.py")

        # template
        template_path = os.path.join(base_dir, main_app_name, 'templates', 'silica', *subfolders, f"{existing_snake_component_name}.html")

        # check if component exists
        component_exists = os.path.exists(component_path)
        template_exists = os.path.exists(template_path)

        # import component
        try:
            dotted_path = f"{main_app_name}.silica.{component_name}.{component_name}"
            component_class = import_string(dotted_path)
            if component_class:
                # check if component has inline_template method
                inline_component: bool = hasattr(component_class, 'inline_template')
        except ImportError:
            inline_component: bool = False

        if component_exists or template_exists:
            if skip_confirmation:
                # Remove without confirmation
                if component_exists:
                    os.remove(component_path)
                    if inline_component:
                        self.stdout.write(self.style.WARNING(f"Inline component {component_path} removed."))
                    else:
                        self.stdout.write(self.style.SUCCESS(f"Component {component_path} removed."))
                if template_exists:
                    os.remove(template_path)
                    self.stdout.write(self.style.SUCCESS(f"Template {template_path} removed."))
            else:
                # Confirmation for each cases
                if component_exists and template_exists:
                    confirmation_message = f"Are you sure you want to remove the component {component_path} and its template {template_path}?"
                elif component_exists:
                    if inline_component:
                        confirmation_message = f"Are you sure you want to remove the inline component {component_path}?"
                    else:
                        confirmation_message = f"Are you sure you want to remove the component {component_path}?"
                elif template_exists:
                    confirmation_message = f"Are you sure you want to remove the template {template_path}?"

                self.stdout.write(self.style.WARNING(confirmation_message))
                confirmation = input("Y/n: ")
                if confirmation.lower() == "y":
                    if component_exists:
                        os.remove(component_path)
                        self.stdout.write(self.style.SUCCESS(f"Component {component_path} removed."))
                    if template_exists:
                        os.remove(template_path)
                        self.stdout.write(self.style.SUCCESS(f"Template {template_path} removed."))
        else:
            self.stdout.write(self.style.ERROR(f"Component {component_path} and template {template_path} not found."))


from urllib.parse import urlparse, parse_qs

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By

from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from silica.tests.urls import urlpatterns as silica_test_url_patterns
from silica.urls import urlpatterns as silica_urlpatterns

urlpatterns = silica_urlpatterns + silica_test_url_patterns


class SilicaBrowserTestCase(StaticLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        CHROME_PATH = '/usr/bin/chromium'
        CHROMEDRIVER_PATH = '/usr/bin/chromedriver'

        options = Options()

        # Basic required options
        options.add_argument('--headless=new')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')

        # Connection stability improvements
        options.add_argument('--disable-gpu')
        options.add_argument('--disable-features=NetworkService')
        options.add_argument('--disable-features=VizDisplayCompositor')
        options.add_argument('--force-device-scale-factor=1')

        # Memory & process management
        options.add_argument('--single-process')  # Important for connection stability
        options.add_argument('--disable-breakpad')
        options.add_argument('--disable-extensions')

        # Browser window settings
        options.add_argument('--window-size=1920,1080')
        options.add_argument('--start-maximized')

        # Logging
        options.add_argument('--log-level=3')  # FATAL only
        options.set_capability('goog:loggingPrefs', {'browser': 'ALL'})

        # Additional stability settings
        options.add_argument('--disable-backgrounding-occluded-windows')
        options.add_argument('--disable-renderer-backgrounding')
        options.add_argument('--disable-site-isolation-trials')

        options.binary_location = CHROME_PATH

        service = Service(
            CHROMEDRIVER_PATH,
            service_args=['--verbose']
        )

        cls.selenium = webdriver.Chrome(options=options, service=service)
        cls.selenium.set_page_load_timeout(10)


    def get_query_param(self, param):
        url = self.selenium.current_url
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)
        return query_params.get(param, [None])[0]

    def get_url(self):
        return self.selenium.current_url

    def get_page_source(self):
        return self.selenium.page_source

    def get_console_log(self):
        return self.selenium.get_log('browser')

    def assertConsoleLogContains(self, text):
        logs = self.get_console_log()
        for log in logs:
            if text in log['message']:
                return self.assertTrue(True)
        return self.assertTrue(False, f'"{text}" is not seen in the console logs: {logs}')

    def assertConsoleLogNotContains(self, text):
        logs = self.get_console_log()
        for log in logs:
            if text in log['message']:
                return self.assertTrue(False, f'"{text}" is seen in the console logs: {logs}')
        return self.assertTrue(True)

    def assertHtmlContains(self, text):
        assert text in self.selenium.page_source, f'"{text}" is not seen in the page source'
        return self

    def assertHtmlNotContains(self, text):
        assert text not in self.selenium.page_source, f'"{text}" is seen in the page source'
        return self

    def findElementById(self, element_id):
        return self.selenium.find_element(By.ID, element_id)

    def assertElementIsVisible(self, element):
        self.assertTrue(element.is_displayed())
        return self

    def assertIdIsVisible(self, element_id):
        element = self.findElementById(element_id)
        self.assertTrue(element.is_displayed())
        return self

    def assertIdIsNotVisible(self, element_id):
        element = self.findElementById(element_id)
        self.assertFalse(element.is_displayed())
        return self

    def assertSee(self, text):
        page_text = self.selenium.find_element(By.TAG_NAME, 'body').text
        self.assertIn(text, page_text, f"'{text}' was not found in the visible rendered content")
        return self

    def assertDontSee(self, text):
        page_text = self.selenium.find_element(By.TAG_NAME, 'body').text
        self.assertNotIn(text, page_text, f"'{text}' was found in the visible rendered content, but it shouldn't be there")
        return self

    @classmethod
    def tearDownClass(cls):
        cls.selenium.quit()
        super().tearDownClass()
        pass




from django.test import TestCase, override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaTestCase import SilicaTest
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns

TestView = create_test_view(f"""{{% silica '{__name__}.LazyLoading' %}}""")

urlpatterns = silica_urlpatterns + [
    path("lazy", TestView.as_view()),
]


class LazyLoading(Component):
    is_mounted: bool = False

    def mount(self):
        self.is_mounted = True

    def inline_template(self):
        return """
            <div>   
                <h1>I will be lazy loaded!</h1>
            </div>
        """

    def inline_placeholder(self):
        return """
            <div>
                <h1>Loading...</h1>
            </div>
        """


class TestLazyLoading(TestCase):
    def test_mount_is_not_called(self):
        (
            SilicaTest(component=LazyLoading, lazy=True)
            .assertSet("is_mounted", False)
        )

    def test_mount_is_called_on_non_lazy(self):
        (
            SilicaTest(component=LazyLoading, lazy=False)
            .assertSet("is_mounted", True)
        )

    def test_placeholder_is_seen(self):
        (
            SilicaTest(component=LazyLoading, lazy=True)
            .assertSee("Loading...")
        )

    def test_placeholder_is_not_seen_on_non_lazy(self):
        (
            SilicaTest(component=LazyLoading, lazy=False)
            .assertDontSee("Loading...")
        )

    def test_silica_lazy_attribute_is_present_on_initial_render(self):
        (
            SilicaTest(component=LazyLoading, lazy=True)
            .assertSee("silica:lazy")
        )

    def test_silica_lazy_attribute_is_not_present_on_initial_render_on_non_lazy(self):
        (
            SilicaTest(component=LazyLoading, lazy=False)
            .assertDontSee("silica:lazy")
        )


@override_settings(ROOT_URLCONF=__name__)
class BrowserTestLazyLoading(SilicaBrowserTestCase):
    def test_lazy_loaded_component_loads(self):
        self.selenium.get(self.live_server_url + '/lazy')

        self.assertHtmlContains("I will be lazy loaded!")


import time

from django.test import override_settings
from django.urls import path

from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.DebounceComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("debounce", TestView.as_view()),
]


class DebounceComponent(Component):
    text = ''
    updated_count = 0
    
    def updated_text(self, value):
        self.text = value
        self.updated_count += 1
        
    def change_text(self):
        self.text = 'changed'

    def inline_template(self):
        return """
            <div>
                default: "{{ text }}"
                custom 600ms: "{{ text }}"
                <input type="text" id="text" silica:model.debounce="text">              
                <input type="text" id="text_600" silica:model.debounce.600ms="text">
                <button id="change_text_btn" silica:click="change_text">Change text</button>              
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class DebounceTestCase(SilicaBrowserTestCase):

    def test_that_debounce_is_delaying_send(self):
        self.selenium.get(self.live_server_url + '/debounce')
        
        element = self.selenium.find_element(By.ID, 'text')
        element.send_keys('a')
        
        time.sleep(0.2)
        
        self.assertHtmlContains('default: ""')

    def test_that_debounce_resolves_after_default_timeout_of_300ms(self):
        self.selenium.get(self.live_server_url + '/debounce')
        
        element = self.selenium.find_element(By.ID, 'text')
        element.send_keys('a')
        
        time.sleep(0.4)
        
        self.assertHtmlContains('default: "a"')

    def test_that_debounce_waits_for_a_custom_timeout(self):
        self.selenium.get(self.live_server_url + '/debounce')
        
        element = self.selenium.find_element(By.ID, 'text_600')
        element.send_keys('b')
        
        time.sleep(0.4)
        
        self.assertHtmlNotContains('custom 600ms: "b"')

    def test_that_debounce_resolves_after_a_custom_timeout(self):
        self.selenium.get(self.live_server_url + '/debounce')
        
        element = self.selenium.find_element(By.ID, 'text_600')
        element.send_keys('b')
        
        time.sleep(0.65)
        
        self.assertHtmlContains('custom 600ms: "b"')
        
    def test_that_inputs_with_debounce_modifier_are_updated_from_backend_update(self):
        self.selenium.get(self.live_server_url + '/debounce')
        
        btn = self.selenium.find_element(By.ID, 'change_text_btn')
        btn.click()
        
        time.sleep(0.2)
        
        input = self.selenium.find_element(By.ID, 'text')
        self.assertEqual(input.get_attribute('value'), 'changed')
        self.assertHtmlContains('default: "changed"')
        
  


from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class Properties(Component):
    foo = "bar"

    @property
    def ding(self):
        return 'dong'

    def inline_template(self):
        return """
            <div>
            foo: {{ foo }}
            ding: {{ ding }}
            </div>
        """


class PropertiesTestCase(SilicaTestCase):
    def test_can_set_properties(self):
        (
            SilicaTest(component=Properties)
            .assertSet("foo", "bar")
            .assertSee("bar")
            .set("foo", "boo")
            .assertSet("foo", "boo")
            .assertSee("boo")
        )

    def test_python_property(self):
        (
            SilicaTest(component=Properties)
            .assertSee("ding: dong")
        )



import time

from django.test import override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.SilicaInitComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("silica-init", TestView.as_view()),
]


class SilicaInitComponent(Component):
    instant_data = "I'm instant"
    mounted_data: str = None
    init_data: str = ''

    def mount(self):
        self.mounted_data = "I'm from mounted"

    def slow_method(self):
        time.sleep(0.4)
        self.init_data = "I'm from an init method"

    def inline_template(self):
        return """
            <div silica:init="slow_method">
                <style>.red { color: red }</style>
                <style>.hidden { display: none }</style>
            
                {{ instant_data }}
                {{ mounted_data }}
                {{ init_data }}       

                <div id="loading-message" silica:loading>I am loading no matter what mate</div>
                <div id="initiating-message" silica:loading.init>I am an initiating message</div>    
                <div id="proceeding-message" silica:loading.proceeding>I am a proceeding message</div>
                <div id="class-message" silica:loading.class="red">I am red whilst loading</div>
                
                <div id="loading-hidden" silica:loading.hidden>I am hidden whilst loading</div> 
                <div id="initiating-hidden" silica:loading.init.hidden>I am hidden whilst initiating</div> 
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class SilicaInitTestCase(SilicaBrowserTestCase):

    def test_init_functionality(self):
        self.selenium.get(self.live_server_url + '/silica-init')

        self.assertHtmlContains("I'm instant")
        self.assertHtmlContains("I'm from mounted")
        self.assertHtmlNotContains("I'm from an init method")

        time.sleep(0.5)

        self.assertHtmlContains("I'm from an init method")

    def test_initiating_loader(self):
        self.selenium.get(self.live_server_url + '/silica-init')

        self.assertIdIsVisible('loading-message')
        self.assertIdIsVisible('initiating-message')
        self.assertIdIsVisible('class-message')
        self.assertHtmlContains(' class="red"')
        self.assertIdIsNotVisible('proceeding-message')

        time.sleep(0.5)

        self.assertIdIsNotVisible('loading-message')
        self.assertIdIsNotVisible('initiating-message')
        self.assertIdIsVisible('class-message')
        self.assertHtmlNotContains(' class="red"')
        
    





import time

from django.test import override_settings
from django.urls import path

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


TestView = create_test_view(f"""{{% silica '{__name__}.MorphTestComponent' %}}""")

urlpatterns = silica_urlpatterns + [
    path("morph", TestView.as_view()),
]


class MorphTestComponent(Component):
    something = "hello"
    items = [1,2,3,4]

    def update_something(self):
        self.something = "world"

    def inline_template(self):
        return """
            <div>
                {{ something }}
                <div x-data="{'expanded': false}">
                    {% for item in items %}
                        <div id="item_{{ item }}" x-show="{% if forloop.counter > 2 %}expanded{% else %}true{% endif %}">
                            item: "{{ item }}"
                        </div>
                    {% endfor %}

                    <button @click="expanded = !expanded" id="toggle"><span x-text="expanded ? 'Hide' : 'Show more'"></span></button>                        
                </div>  
                <button silica:click.prevent="update_something" id="update_something">Update</button>           
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class MorphTestCase(SilicaBrowserTestCase):

    def test_rerenders_doesnt_break_alpine_component(self):
        self.selenium.get(self.live_server_url + '/morph')

        # First let's check we have a working alpine js component
        self.assertIdIsVisible('item_1')
        self.assertIdIsVisible('item_2')
        self.assertIdIsNotVisible('item_3')
        self.assertIdIsNotVisible('item_4')

        self.findElementById('toggle').click()

        self.assertIdIsVisible('item_1')
        self.assertIdIsVisible('item_2')
        self.assertIdIsVisible('item_3')
        self.assertIdIsVisible('item_4')

        # Now let's check that we can update the component without breaking it
        self.findElementById('update_something').click()
        time.sleep(0.2)
        self.assertHtmlContains("world")

        self.assertIdIsVisible('item_1')
        self.assertIdIsVisible('item_2')
        self.assertIdIsVisible('item_3')
        self.assertIdIsVisible('item_4')

        self.findElementById('toggle').click()

        self.assertIdIsVisible('item_1')
        self.assertIdIsVisible('item_2')
        self.assertIdIsNotVisible('item_3')
        self.assertIdIsNotVisible('item_4')





import logging
import time

from django.test import override_settings
from django.urls import path
from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns


class ComponentWithError(Component):    
    def some_function(self):
        raise ValueError('Something went wrong')

    def inline_template(self):
        return """
            <div silica:init="some_function">
            </div>
        """


TestView = create_test_view(f"""
    {{% silica '{__name__}.ComponentWithError' %}}
    """)


urlpatterns = silica_urlpatterns + [
    path("component-with-error/", TestView.as_view()),
]


@override_settings(ROOT_URLCONF=__name__, DEBUG=True)
class ErrorModalBrowserTestCase(SilicaBrowserTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Store the original logging levels
        cls.original_log_levels = {
            'django.request': logging.getLogger('django.request').level,
            'django.server': logging.getLogger('django.server').level,
        }
        # Set logging levels to suppress expected errors
        logging.getLogger('django.request').setLevel(logging.CRITICAL)
        logging.getLogger('django.server').setLevel(logging.CRITICAL)

    @classmethod
    def tearDownClass(cls):
        # Restore original logging levels
        for logger_name, level in cls.original_log_levels.items():
            logging.getLogger(logger_name).setLevel(level)
        super().tearDownClass()

    def test_error_modal_shows_in_development(self):
        with self.settings(DEBUG=True):
            self.selenium.get(self.live_server_url + '/component-with-error')
            time.sleep(0.4)
            self.assertTrue("error-modal" in self.selenium.page_source)
            # self.selenium.save_screenshot('error_modal.png')
            iframe = self.selenium.find_element(By.ID, 'error-content-frame')
            self.selenium.switch_to.frame(iframe)
            self.assertTrue("Something went wrong" in self.selenium.page_source)
            
    def test_error_modal_shows_in_production(self):
        with self.settings(DEBUG=False):
            self.selenium.get(self.live_server_url + '/component-with-error')
            time.sleep(2)
            self.assertTrue("error-modal" in self.selenium.page_source)
            # self.selenium.save_screenshot('error_modal.png')
            iframe = self.selenium.find_element(By.ID, 'error-content-frame')
            self.selenium.switch_to.frame(iframe)
            self.assertTrue("Server Error (500)" in self.selenium.page_source)
                



from pathlib import Path
from silica.tests.SilicaTestCase import SilicaTestCase
from django.core.management import call_command
from django.conf import settings
import importlib.util


def import_from_path(path: Path, object_name: str):
    # Ensure the path is absolute
    if not path.is_absolute():
        path = path.resolve()

    # Get the module name and import path
    module_name = path.stem
    spec = importlib.util.spec_from_file_location(module_name, path)

    # Load the module from the spec
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Return the desired object from the module
    return getattr(module, object_name)


class ComponentTagTestCase(SilicaTestCase):
    # setup create files of self.files list
    def setUp(self):
        self.files = []

    # teardown delete files of self.files list
    def tearDown(self):
        for file in self.files:
            if file.exists():
                file.unlink()

    def test_template_and_component_name_are_snake_case(self):
        call_command('silica', 'create', "TestUniqueSideBar", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # component
        pascal_component_path = main_app_dir / 'silica' / "TestUniqueSideBar.py"
        self.files.append(pascal_component_path)
        snake_component_path = main_app_dir / 'silica' / "test_unique_side_bar.py"
        self.files.append(snake_component_path)

        # template
        pascal_template_path = main_app_dir / 'templates' / 'silica' / "TestUniqueSideBar.html"
        self.files.append(pascal_template_path)
        snake_template_path = main_app_dir / 'templates' / 'silica' / "test_unique_side_bar.html"
        self.files.append(snake_template_path)

        self.assertTrue(pascal_component_path.exists())
        self.assertTrue(snake_template_path.exists())


    def test_silica_create_with_nested_subfolders(self):
        call_command('silica', 'create', "folder1.folder2.folder3.NewComponent", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # Component in nested subfolders
        component_path = main_app_dir / 'silica' / 'folder1' / 'folder2' / 'folder3' / "NewComponent.py"
        self.files.append(component_path)

        # Template in nested subfolders
        template_path = main_app_dir / 'templates' / 'silica' / 'folder1' / 'folder2' / 'folder3' / "new_component.html"
        self.files.append(template_path)

        incorrect_component_path = main_app_dir / 'silica' / 'folder1.folder2.folder3' / "NewComponent.py"
        self.files.append(incorrect_component_path)
        incorrect_template_path = main_app_dir / 'templates' / 'silica' / 'folder1.folder2.folder3' / "new_component.html"
        self.files.append(incorrect_template_path)

        self.assertTrue(component_path.exists())
        self.assertTrue(template_path.exists())


    def test_silica_template_name_in_component_definition(self):
        call_command('silica', 'create', "folder1.folder2.folder3.NewComponent", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # Component in nested subfolders
        component_path = main_app_dir / 'silica' / 'folder1' / 'folder2' / 'folder3' / "NewComponent.py"
        self.files.append(component_path)

        # Template in nested subfolders
        template_path = main_app_dir / 'templates' / 'silica' / 'folder1' / 'folder2' / 'folder3' / "new_component.html"
        self.files.append(template_path)

        self.assertTrue(component_path.exists())
        self.assertTrue(template_path.exists())

        Component = import_from_path(component_path, "NewComponent")
        self.assertEqual(Component.template_name, "silica/folder1/folder2/folder3/new_component.html")


    def test_command_allows_kebab_case(self):
        call_command('silica', 'create', "some.Sub-folder.some-component", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # component
        wrong_component_path = main_app_dir / 'silica' / "some/Sub-folder/some-component.py"
        expected_component_path = main_app_dir / 'silica' / "some/sub-folder/SomeComponent.py"
        self.files.append(expected_component_path)
        self.files.append(wrong_component_path)

        # template
        wrong_template_path = main_app_dir / 'templates' / 'silica' / "some/Sub-folder/some-component.html"
        expected_template_path = main_app_dir / 'templates' / 'silica' / "some/sub-folder/some_component.html"
        self.files.append(expected_template_path)
        self.files.append(wrong_template_path)

        self.assertTrue(expected_component_path.exists())
        self.assertTrue(expected_template_path.exists())


    def test_command_allows_pascal_case(self):
        call_command('silica', 'create', "some.Sub-folder.SomeComponent", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # component
        wrong_component_path = main_app_dir / 'silica' / "some/Sub-folder/some-component.py"
        expected_component_path = main_app_dir / 'silica' / "some/sub-folder/SomeComponent.py"
        self.files.append(expected_component_path)
        self.files.append(wrong_component_path)

        # template
        wrong_template_path = main_app_dir / 'templates' / 'silica' / "some/Sub-folder/some-component.html"
        expected_template_path = main_app_dir / 'templates' / 'silica' / "some/sub-folder/some_component.html"
        self.files.append(expected_template_path)
        self.files.append(wrong_template_path)

        self.assertTrue(expected_component_path.exists())
        self.assertTrue(expected_template_path.exists())


    def test_single_word_component_is_capitalized(self):
        call_command('silica', 'create', "some.folder.component", "--silent")
        main_app_dir = Path(settings.BASE_DIR) / "example_project"

        # component
        wrong_component_path = main_app_dir / 'silica' / "some/folder/component.py"
        expected_component_path = main_app_dir / 'silica' / "some/folder/Component.py"
        self.files.append(expected_component_path)
        self.files.append(wrong_component_path)

        # template
        expected_template_path = main_app_dir / 'templates' / 'silica' / "some/folder/component.html"
        self.files.append(expected_template_path)

        self.assertTrue(expected_component_path.exists())
        self.assertTrue(expected_template_path.exists())



pass

"""
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.redis.RedisCache",
        "LOCATION": f"redis://default:{CFG_REDIS_PASSWORD}@{CFG_REDIS_HOST}:{CFG_REDIS_PORT}",
        "TIMEOUT": 1 * 60 * 60 * 24,
        # "TIMEOUT": 1,
    }
}"""

# test an exception is raised when cache expires
# test when cache expires, we get an alert telling us we must refresh page


from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class Lifecycle(Component):
    called_mount = 0
    called_updated = None
    called_updated_property = None

    property = None
    dict_property = {
        "foo": "bar",
        "baz": {
            "jim": "bob"
        }
    }

    def mount(self):
        self.called_mount += 1

    def updated_property(self, value):
        self.called_updated_property = value

    def updated(self, prop, value):
        self.called_updated = f"{prop}={value}"

    def updated_dict_property(self, key, value):
        self.called_updated_property = f"{key}={value}"

    def inline_template(self):
        return """
            <div>
                hi!
            </div>        
        """


class LifecycleTestCase(SilicaTestCase):
    def test_mount_is_called_once(self):
        (
            # Initial request
            SilicaTest(component=Lifecycle)
            .assertSet("called_mount", 1)
            .set("property", "test")
            .assertSet("called_updated_property", "test")
            .assertSet("called_updated", "property=test")
            .set("dict_property.foo", "baz")
            .assertSet("called_updated_property", "foo=baz")
            .assertSet("called_updated", "dict_property.foo=baz")
            .set("dict_property.baz.jim", "jon")
            .assertSet("called_updated_property", "baz.jim=jon")
            .assertSet("called_updated", "dict_property.baz.jim=jon")
        )




# Enhanced caching for improved performance on repetitive parsing tasks
import ast


def kwarg_as_dict(kwarg_str: str) -> dict:
    """
    Parses a string representation of a keyword argument into a dictionary with the value
    appropriately typed.

    Args:
        kwarg_str: A string representing a keyword argument, e.g., "change_prop='True'".

    Returns:
        A dictionary with the keyword as the key and the evaluated Python literal as the value.

    Raises:
        ValueError: If the input string cannot be parsed into a valid keyword argument.
    """
    # Attempt to split the input string into a key and value part
    try:
        key, value_str = kwarg_str.split('=', 1)
        key = key.strip()
        # Using ast.literal_eval to safely evaluate the value string to a Python literal
        value = ast.literal_eval(value_str.strip())
    except ValueError as e:
        raise ValueError(f"Invalid kwarg string '{kwarg_str}': {e}")
    except SyntaxError as e:
        raise ValueError(f"Syntax error in value for '{kwarg_str}': {e}")

    return {key: value}

def pascal_to_snake(s):
    """
    Convert a PascalCase string to snake_case.

    :param s: String in PascalCase format
    :return: String converted to snake_case format
    """
    if not s:
        return ""

    snake_case = [s[0].lower()]
    for char in s[1:]:
        if char.isupper():
            snake_case.append('_')
            snake_case.append(char.lower())
        else:
            snake_case.append(char)

    return ''.join(snake_case)


from silica import Component
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest


class EventsCaller(Component):
    def mount(self):
        self.emit('event_name', {'something': 1})

    def set_event(self):
        self.emit('another_event', {'else': 2})

    def method_without_event_emitting(self):
        pass

    def inline_template(self):
        return """
            <div>
                I call events!
            </div>
        """


class ComponentWithListener(Component):
    received_event_payload = {}

    def some_event_name(self, payload):
        self.received_event_payload = payload

    def inline_template(self):
        return """
            <div>
                Event called:
                <div>{{ received_event_name }}</div>
                <div>{{ received_event_payload }}</div>
            </div>
        """


class LazyComponentEventReceiver(Component):
    received_event_payload = {}

    def some_event_name(self, payload):
        self.received_event_payload = payload

    def inline_template(self):
        return """
            <div>
                Event called:
                <div>{{ received_event_name }}</div>
                <div>{{ received_event_payload }}</div>
            </div>
        """


class EventsTest(SilicaTestCase):
    def test_event_is_called_from_mount(self):
        (
            SilicaTest(component=EventsCaller)
            .assertEventEmitted('event_name', {'something': 1})
        )

    def test_event_is_called_from_message(self):
        request = (
            SilicaTest(component=EventsCaller)
            .call('set_event')
            .assertEventEmitted('another_event', {'else': 2})
        )

        (request
            .call('method_without_event_emitting')
            .assertEventNotEmitted('another_event')
         )

    def test_an_event_listener_is_called_and_payload_received(self):
        (
            SilicaTest(component=ComponentWithListener)
            .emit('some_event_name', {'something': 1})
            .assertSee("{'something': 1}")
        )

    def test_emitting_event_to_one_component_name(self):
        (
            SilicaTest(component=ComponentWithListener)
            .emit_to('some_event_name', {'something': 1})
            .assertSee("{'something': 1}")
        )

    def test_emitting_event_to_one_component_name_only(self):
        # Needs browser testing
        pass

    def test_emitting_event_to_multiple_components_with_same_name(self):
        # Needs browser testing
        pass

    def test_emitting_events_to_lazy_components(self):
        # Not really able to test properly, need browser testing
        (
            SilicaTest(component=LazyComponentEventReceiver, lazy=True)
            .emit('some_event_name', {'something': 1})
            .assertSee("{'something': 1}")
        )

    def test_emitting_events_from_lazy_components(self):
        pass


from pathlib import Path
from silica.tests.SilicaTestCase import SilicaTestCase
import importlib.util
from django.test import Client


def import_from_path(path: Path, object_name: str):
    # Ensure the path is absolute
    if not path.is_absolute():
        path = path.resolve()

    # Get the module name and import path
    module_name = path.stem
    spec = importlib.util.spec_from_file_location(module_name, path)

    # Load the module from the spec
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Return the desired object from the module
    return getattr(module, object_name)

class ComponentTagTestCase(SilicaTestCase):
    # setup create files of self.files list
    def setUp(self):
        self.files = []

    # teardown delete files of self.files list
    def tearDown(self):
        for file in self.files:
            if file.exists():
                file.unlink()


    def test_short_tag_can_be_called(self):
        client = Client()
        response = client.get("/silica/tests/component-tag-test")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "I'm component called with short name!")


    def test_components_in_subfolders_can_be_called(self):
        client = Client()
        response = client.get("/silica/tests/component-subfolder-test")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "I'm component in subfolder!")



from io import StringIO
import os
from pathlib import Path
import random
from silica.utils import pascal_to_snake
from django.core.management.base import BaseCommand, CommandError

def snake_to_camel(snake_str):
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


class Command(BaseCommand):
    help = 'Creates a new Silica component'
    quotes = [
        'You are in control of your own destiny.',
        'Create something awesome.',
        'Carpe diem.',
    ]
    TEMPLATE_FOR_CLASS = """from silica import Component


class {class_name}(Component):
    template_name = "{template_name}"
    
"""

    TEMPLATE_FOR_INLINE_CLASS = """from silica import Component


class {class_name}(Component):

    def inline_template(self):
        return \"\"\"
            <div>
            {quote}
            </div>
        \"\"\"
    # Your class implementation here
"""

    TEMPLATE_FOR_TEMPLATE = f"""
{{# {random.choice(quotes)} #}}
"""

    def add_arguments(self, parser):
        parser.add_argument('component_name', type=str)
        parser.add_argument('--inline', action='store_true', help='Use inline template')
        parser.add_argument('--app', type=str, help='Specify the Django app name')
        parser.add_argument('--silent', action='store_true', help='Run the command in silent mode without output')

    def handle(self, *args, **options):
        component_name: str = options['component_name']
        use_inline_template = options['inline']
        app_name = options['app']
        silent = options['silent']


        if silent:
            self.stdout = StringIO()

        subfolders: list[str] = []

        if '.' in component_name:
            *subfolders, component_name = component_name.split(".") # list packing

        subfolders = [subfolder.lower() for subfolder in subfolders]
        component_name = component_name.replace('-', '_')

        if "_" in component_name:
            component_name = snake_to_camel(component_name)
        else:
            # Capitalize the first letter
            component_name = component_name[0].upper() + component_name[1:]
        
        # Convert component name to snake_case for the template
        snake_component_name = pascal_to_snake(component_name)


        # Choose the template based on the --inline flag
        if use_inline_template:
            component_template = self.TEMPLATE_FOR_INLINE_CLASS.format(class_name=component_name, quote=random.choice(self.quotes))
            write_template = False
        else:
            write_template = True
            template_name = Path("silica/") / "/".join(subfolders) / f"{snake_component_name}.html"
            component_template = self.TEMPLATE_FOR_CLASS.format(class_name=component_name, template_name=str(template_name))



        if app_name:
            main_app_name = app_name
        else:
            settings_module = os.environ.get('DJANGO_SETTINGS_MODULE')
            if not settings_module:
                raise CommandError("DJANGO_SETTINGS_MODULE environment variable not set.")
            main_app_name = settings_module.split('.')[0]

        

        # Get the BASE_DIR from Django settings
        from django.conf import settings
        base_dir = getattr(settings, 'BASE_DIR')
        if not base_dir:
            raise CommandError("Unable to determine the BASE_DIR from Django settings.")

        if not os.path.exists(os.path.join(base_dir, main_app_name)):
            raise CommandError(f"App {main_app_name} not found.")

        component_path = os.path.join(base_dir, main_app_name, 'silica', *subfolders, f"{component_name}.py")
        template_path = os.path.join(base_dir, main_app_name, 'templates', 'silica', *subfolders, f"{snake_component_name}.html")

        # Check if files already exist
        if os.path.exists(component_path) or os.path.exists(template_path):
            raise CommandError(f"Component {component_name} already exists.")

        # Create component file
        os.makedirs(os.path.dirname(component_path), exist_ok=True)
        with open(component_path, 'w') as f:
            f.write(component_template)
            self.stdout.write(self.style.SUCCESS(f"Component created at {component_path}"))

        # Create template file
        if write_template:
            os.makedirs(os.path.dirname(template_path), exist_ok=True)
            with open(template_path, 'w') as f:
                f.write(self.TEMPLATE_FOR_TEMPLATE)
                self.stdout.write(self.style.SUCCESS(f"Template created at {template_path}"))


import time

from django.test import override_settings
from django.urls import path
from selenium.webdriver.common.by import By

from silica import Component
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns

TestView = create_test_view(f"""{{% silica '{__name__}.ConcurrencyTest' %}}""")

urlpatterns = silica_urlpatterns + [
    path("concurrency", TestView.as_view()),
]


class ConcurrencyTest(Component):
    ui_value: str = None

    def slow_first(self):
        time.sleep(2)
        self.ui_value = "data_from_request_1"

    def slow_second(self):
        time.sleep(1)
        self.ui_value = "data_from_request_2"

    def inline_template(self):
        return """
            <div>
                <button id="slow_request_first" silica:click.prevent="slow_first">Slow</button>
                <button id="quick_request_second" silica:click.prevent="slow_second">Quick</button>
                <div id="updated_element">{{ ui_value }}</div>
            </div>
        """


@override_settings(ROOT_URLCONF=__name__)
class ConcurrencyTestCase(SilicaBrowserTestCase):

    def test_a_slower_earlier_request_doesnt_overwrite_a_later_quicker_request(self):
        # Navigate to the page
        self.selenium.get(self.live_server_url + '/concurrency')

        # Trigger the first request
        self.selenium.find_element(By.ID, 'slow_request_first').click()

        # Without waiting, trigger the second request
        self.selenium.find_element(By.ID, 'quick_request_second').click()

        # Give time for the requests to complete
        time.sleep(3)

        # Wait for a UI element that gets updated by the responses
        updated_element = self.selenium.find_element(By.ID, 'updated_element')

        # Check that the updated_element contains data from the 2nd request ONLY.
        self.assertNotIn('data_from_request_1', updated_element.text)
        self.assertIn('data_from_request_2', updated_element.text)



import time

from django.test import RequestFactory, override_settings
from django.contrib.auth.models import AnonymousUser

from silica import Component
from silica.tests.utils import create_test_view
from silica.urls import urlpatterns as silica_urlpatterns, path
from silica.tests.SilicaTestCase import SilicaTestCase, SilicaTest
from silica.tests.SilicaBrowserTestCase import SilicaBrowserTestCase

from selenium.webdriver.common.by import By

TestView = create_test_view(f"""
    {{% silica '{__name__}.QueryParamsTestComponent' %}}
    """)

urlpatterns = silica_urlpatterns + [
    path("query-params/", TestView.as_view()),
]


class QueryParamsTestComponent(Component):
    property_1 = "foo1"
    property_2 = "foo2"
    property_3 = "foo3"
    list_property = []

    query_params = [
        "property_1",
        {"param": "property_2"},
        {"param": "property_3", "as": "p3"},
        "list_property"
    ]

    def inline_template(self):
        return """
            <div>
                <p>
                    property1: "{{ property_1 }}"
                    <button silica:click.prevent="property_1 = 'foo1'" id="set_to_default">Set</button>
                </p>
                <p>property2: "{{ property_2 }}"</p>
                <p>property3: "{{ property_3 }}"</p>
                
                <p>{% for item in list_property %}{{ item }}{% endfor %}</p>
                <p><label><input type="checkbox" silica:model="list_property" value="foo"> foo</label></p>
                <p><label><input type="checkbox" silica:model="list_property" value="bar"> bar</label></p>
                <p><label><input type="checkbox" silica:model="list_property" value="baz"> baz</label></p> 
            </div>
        """


class QueryParamTests(SilicaTestCase):
    def test_query_params_alt_formats(self):
        (
            SilicaTest(component=QueryParamsTestComponent)
            .assertSet("property_2", "foo2")
            .assertSee("foo2")
            .assertSet("property_3", "foo3")
            .assertSee("foo3")
        )

        request = RequestFactory().get("/?property_2=bar")
        request.user = AnonymousUser()

        (
            SilicaTest(component=QueryParamsTestComponent, request=request)
            .assertSet("property_2", "bar")
            .assertSee("bar")
        )

    def test_query_params_can_be_set(self):
        (
            SilicaTest(component=QueryParamsTestComponent)
            .assertSet("property_1", "foo1")
            .assertSee("foo1")
        )

        request = RequestFactory().get("/?property_1=bar")
        request.user = AnonymousUser()

        (
            SilicaTest(component=QueryParamsTestComponent, request=request)
            .assertSet("property_1", "bar")
            .assertSee("bar")
        )

    def test_aliased_query_params(self):
        (
            SilicaTest(component=QueryParamsTestComponent)
            .assertSet("property_3", "foo3")
            .assertSee("foo3")
        )

        request = RequestFactory().get("/?p3=bar")
        request.user = AnonymousUser()

        (
            SilicaTest(component=QueryParamsTestComponent, request=request)
            .assertSet("property_3", "bar")
            .assertSee("bar")
        )


@override_settings(ROOT_URLCONF=__name__)
class QueryParamBrowserTests(SilicaBrowserTestCase):
    def test_query_params_are_removed_for_default_values(self):
        self.selenium.get(self.live_server_url + "/query-params?property_1=bar")

        self.assertTrue(self.get_query_param("property_1") == "bar", "Param not set in url")

        self.selenium.find_element(By.ID, 'set_to_default').click()
        time.sleep(0.2)

        self.assertTrue(self.get_query_param("property_1") is None, "Param not cleared when set to default")


    def test_list_query_params_are_in_url_correctly(self):
        self.selenium.get(self.live_server_url + "/query-params")

        self.selenium.find_element(By.XPATH, "//input[@value='foo']").click()
        self.selenium.find_element(By.XPATH, "//input[@value='bar']").click()
        self.selenium.find_element(By.XPATH, "//input[@value='baz']").click()

        time.sleep(0.2)

        self.assertTrue("list_property=foo&list_property=bar&list_property=baz" in self.get_url())